# SPDX-License-Identifier: MIT
from __future__ import annotations

import types
from functools import total_ordering
from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    Dict,
    Iterator,
    List,
    NamedTuple,
    NoReturn,
    Optional,
    Type,
    TypeVar,
)

if TYPE_CHECKING:
    from typing_extensions import Self

__all__ = (
    "Enum",
    "ChannelType",
    "MessageType",
    "SpeakingState",
    "VerificationLevel",
    "ContentFilter",
    "Status",
    "DefaultAvatar",
    "AuditLogAction",
    "AuditLogActionCategory",
    "UserFlags",
    "ActivityType",
    "NotificationLevel",
    "TeamMembershipState",
    "TeamMemberRole",
    "WebhookType",
    "ExpireBehaviour",
    "ExpireBehavior",
    "StickerType",
    "StickerFormatType",
    "InviteType",
    "InviteTarget",
    "VideoQualityMode",
    "ComponentType",
    "ButtonStyle",
    "TextInputStyle",
    "SelectDefaultValueType",
    "StagePrivacyLevel",
    "InteractionType",
    "InteractionResponseType",
    "NSFWLevel",
    "OptionType",
    "ApplicationCommandType",
    "ApplicationCommandPermissionType",
    "PartyType",
    "GuildScheduledEventEntityType",
    "GuildScheduledEventStatus",
    "GuildScheduledEventPrivacyLevel",
    "ThreadArchiveDuration",
    "WidgetStyle",
    "Locale",
    "AutoModTriggerType",
    "AutoModEventType",
    "AutoModActionType",
    "ThreadSortOrder",
    "ThreadLayout",
    "Event",
    "ApplicationRoleConnectionMetadataType",
    "OnboardingPromptType",
    "SKUType",
    "EntitlementType",
    "SubscriptionStatus",
    "PollLayoutType",
    "VoiceChannelEffectAnimationType",
    "MessageReferenceType",
)


class _EnumValueBase(NamedTuple):
    if TYPE_CHECKING:
        _cls_name: ClassVar[str]

    name: str
    value: Any

    def __repr__(self) -> str:
        return f"<{self._cls_name}.{self.name}: {self.value!r}>"

    def __str__(self) -> str:
        return f"{self._cls_name}.{self.name}"


@total_ordering
class _EnumValueComparable(_EnumValueBase):
    def __eq__(self, other: object) -> bool:
        return isinstance(other, self.__class__) and self.value == other.value

    def __lt__(self, other: object) -> bool:
        return isinstance(other, self.__class__) and self.value < other.value


def _create_value_cls(name: str, comparable: bool) -> Type[_EnumValueBase]:
    parent = _EnumValueComparable if comparable else _EnumValueBase
    return type(f"{parent.__name__}_{name}", (parent,), {"_cls_name": name})  # type: ignore


def _is_descriptor(obj):
    return hasattr(obj, "__get__") or hasattr(obj, "__set__") or hasattr(obj, "__delete__")


class EnumMeta(type):
    if TYPE_CHECKING:
        __name__: ClassVar[str]
        _enum_member_names_: ClassVar[List[str]]
        _enum_member_map_: ClassVar[Dict[str, Any]]
        _enum_value_map_: ClassVar[Dict[Any, Any]]
        _enum_value_cls_: ClassVar[Type[_EnumValueBase]]

    def __new__(cls, name: str, bases, attrs, *, comparable: bool = False):
        value_mapping = {}
        member_mapping = {}
        member_names = []

        value_cls = _create_value_cls(name, comparable)
        for key, value in list(attrs.items()):
            is_descriptor = _is_descriptor(value)
            if key[0] == "_" and not is_descriptor:
                continue

            # Special case classmethod to just pass through
            if isinstance(value, classmethod):
                continue

            if is_descriptor:
                setattr(value_cls, key, value)
                del attrs[key]
                continue

            try:
                new_value = value_mapping[value]
            except KeyError:
                new_value = value_cls(name=key, value=value)
                value_mapping[value] = new_value
                member_names.append(key)

            member_mapping[key] = new_value
            attrs[key] = new_value

        attrs["_enum_value_map_"] = value_mapping
        attrs["_enum_member_map_"] = member_mapping
        attrs["_enum_member_names_"] = member_names
        attrs["_enum_value_cls_"] = value_cls
        actual_cls = super().__new__(cls, name, bases, attrs)
        value_cls._actual_enum_cls_ = actual_cls  # type: ignore
        return actual_cls

    def __iter__(cls) -> Iterator[Self]:
        return (cls._enum_member_map_[name] for name in cls._enum_member_names_)

    def __reversed__(cls) -> Iterator[Self]:
        return (cls._enum_member_map_[name] for name in reversed(cls._enum_member_names_))

    def __len__(cls) -> int:
        return len(cls._enum_member_names_)

    def __repr__(cls) -> str:
        return f"<enum {cls.__name__}>"

    @property
    def __members__(cls):
        return types.MappingProxyType(cls._enum_member_map_)

    def __call__(cls, value):
        try:
            return cls._enum_value_map_[value]
        except (KeyError, TypeError):
            raise ValueError(f"{value!r} is not a valid {cls.__name__}") from None

    def __getitem__(cls, key):
        return cls._enum_member_map_[key]

    def __setattr__(cls, name: str, value) -> NoReturn:
        raise TypeError("Enums are immutable.")

    def __delattr__(cls, attr) -> NoReturn:
        raise TypeError("Enums are immutable")

    def __instancecheck__(self, instance) -> bool:
        # isinstance(x, Y)
        # -> __instancecheck__(Y, x)
        try:
            return instance._actual_enum_cls_ is self
        except AttributeError:
            return False


if TYPE_CHECKING:
    from enum import Enum
else:

    class Enum(metaclass=EnumMeta):
        @classmethod
        def try_value(cls, value):
            try:
                return cls._enum_value_map_[value]
            except (KeyError, TypeError):
                return value


class ChannelType(Enum):
    """Specifies the type of channel."""

    text = 0
    """A text channel."""
    private = 1
    """A private text channel. Also called a direct message."""
    voice = 2
    """A voice channel."""
    group = 3
    """A private group text channel."""
    category = 4
    """A category channel."""
    news = 5
    """A guild news channel."""
    news_thread = 10
    """A news thread.

    .. versionadded:: 2.0
    """
    public_thread = 11
    """A public thread.

    .. versionadded:: 2.0
    """
    private_thread = 12
    """A private thread.

    .. versionadded:: 2.0
    """
    stage_voice = 13
    """A guild stage voice channel.

    .. versionadded:: 1.7
    """
    guild_directory = 14
    """A student hub channel.

    .. versionadded:: 2.1
    """
    forum = 15
    """A channel of only threads.

    .. versionadded:: 2.5
    """
    media = 16
    """A channel of only threads but with a focus on media, similar to forum channels.

    .. versionadded:: 2.10
    """

    def __str__(self) -> str:
        return self.name


class MessageType(Enum):
    """Specifies the type of :class:`Message`. This is used to denote if a message
    is to be interpreted as a system message or a regular message.
    """

    default = 0
    """The default message type. This is the same as regular messages."""
    recipient_add = 1
    """The system message when a user is added to a group private message or a thread."""
    recipient_remove = 2
    """The system message when a user is removed from a group private message or a thread."""
    call = 3
    """The system message denoting call state, e.g. missed call, started call, etc."""
    channel_name_change = 4
    """The system message denoting that a channel's name has been changed."""
    channel_icon_change = 5
    """The system message denoting that a channel's icon has been changed."""
    pins_add = 6
    """The system message denoting that a pinned message has been added to a channel."""
    new_member = 7
    """The system message denoting that a new member has joined a Guild."""
    premium_guild_subscription = 8
    """The system message denoting that a member has "nitro boosted" a guild."""
    premium_guild_tier_1 = 9
    """The system message denoting that a member has "nitro boosted" a guild and it achieved level 1."""
    premium_guild_tier_2 = 10
    """The system message denoting that a member has "nitro boosted" a guild and it achieved level 2."""
    premium_guild_tier_3 = 11
    """The system message denoting that a member has "nitro boosted" a guild and it achieved level 3."""
    channel_follow_add = 12
    """The system message denoting that an announcement channel has been followed.

    .. versionadded:: 1.3
    """
    guild_stream = 13
    """The system message denoting that a member is streaming in the guild.

    .. versionadded:: 1.7
    """
    guild_discovery_disqualified = 14
    """The system message denoting that the guild is no longer eligible for Server Discovery.

    .. versionadded:: 1.7
    """
    guild_discovery_requalified = 15
    """The system message denoting that the guild has become eligible again for Server Discovery.

    .. versionadded:: 1.7
    """
    guild_discovery_grace_period_initial_warning = 16
    """The system message denoting that the guild has failed to meet the Server
    Discovery requirements for one week.

    .. versionadded:: 1.7
    """
    guild_discovery_grace_period_final_warning = 17
    """The system message denoting that the guild has failed to meet the Server
    Discovery requirements for 3 weeks in a row.

    .. versionadded:: 1.7
    """
    thread_created = 18
    """The system message denoting that a thread has been created. This is only
    sent if the thread has been created from an older message. The period of time
    required for a message to be considered old cannot be relied upon and is up to
    Discord.

    .. versionadded:: 2.0
    """
    reply = 19
    """The system message denoting that the author is replying to a message.

    .. versionadded:: 2.0
    """
    application_command = 20
    """The system message denoting that an application (or "slash") command was executed.

    .. versionadded:: 2.0
    """
    thread_starter_message = 21
    """The system message denoting the message in the thread that is the one that started the
    thread's conversation topic.

    .. versionadded:: 2.0
    """
    guild_invite_reminder = 22
    """The system message sent as a reminder to invite people to the guild.

    .. versionadded:: 2.0
    """
    context_menu_command = 23
    """The system message denoting that a context menu command was executed.

    .. versionadded:: 2.3
    """
    auto_moderation_action = 24
    """The system message denoting that an auto moderation action was executed.

    .. versionadded:: 2.5
    """
    role_subscription_purchase = 25
    """The system message denoting that a role subscription was purchased.

    .. versionadded:: 2.9
    """
    interaction_premium_upsell = 26
    """The system message for an application premium subscription upsell.

    .. versionadded:: 2.8
    """
    stage_start = 27
    """The system message denoting the stage has been started.

    .. versionadded:: 2.9
    """
    stage_end = 28
    """The system message denoting the stage has ended.

    .. versionadded:: 2.9
    """
    stage_speaker = 29
    """The system message denoting a user has become a speaker.

    .. versionadded:: 2.9
    """
    stage_topic = 31
    """The system message denoting the stage topic has been changed.

    .. versionadded:: 2.9
    """
    guild_application_premium_subscription = 32
    """The system message denoting that a guild member has subscribed to an application.

    .. versionadded:: 2.8
    """
    guild_incident_alert_mode_enabled = 36
    """The system message denoting that an admin enabled security actions.

    .. versionadded:: 2.10
    """
    guild_incident_alert_mode_disabled = 37
    """The system message denoting that an admin disabled security actions.

    .. versionadded:: 2.10
    """
    guild_incident_report_raid = 38
    """The system message denoting that an admin reported a raid.

    .. versionadded:: 2.10
    """
    guild_incident_report_false_alarm = 39
    """The system message denoting that a raid report was a false alarm.

    .. versionadded:: 2.10
    """
    poll_result = 46
    """The system message denoting that a poll expired, announcing the most voted answer.

    .. versionadded:: 2.10
    """


class PartyType(Enum):
    """Represents the type of a voice channel activity/application.

    .. deprecated:: 2.9
    """

    poker = 755827207812677713
    """The "Poker Night" activity."""
    betrayal = 773336526917861400
    """The "Betrayal.io" activity."""
    fishing = 814288819477020702
    """The "Fishington.io" activity."""
    chess = 832012774040141894
    """The "Chess In The Park" activity."""
    letter_tile = 879863686565621790
    """The "Letter Tile" activity."""
    word_snack = 879863976006127627
    """The "Word Snacks" activity."""
    doodle_crew = 878067389634314250
    """The "Doodle Crew" activity."""
    checkers = 832013003968348200
    """The "Checkers In The Park" activity.

    .. versionadded:: 2.3
    """
    spellcast = 852509694341283871
    """The "SpellCast" activity.

    .. versionadded:: 2.3
    """
    watch_together = 880218394199220334
    """The "Watch Together" activity, a Youtube application.

    .. versionadded:: 2.3
    """
    sketch_heads = 902271654783242291
    """The "Sketch Heads" activity.

    .. versionadded:: 2.4
    """
    ocho = 832025144389533716
    """The "Ocho" activity.

    .. versionadded:: 2.4
    """
    gartic_phone = 1007373802981822582
    """The "Gartic Phone" activity.

    .. versionadded:: 2.9
    """


# undocumented/internal
class SpeakingState(Enum):
    none = 0
    voice = 1 << 0
    soundshare = 1 << 1
    priority = 1 << 2

    def __str__(self) -> str:
        return self.name

    def __int__(self) -> int:
        return self.value


class VerificationLevel(Enum, comparable=True):
    """Specifies a :class:`Guild`\\'s verification level, which is the criteria in
    which a member must meet before being able to send messages to the guild.

    .. collapse:: operations

        .. versionadded:: 2.0

        .. describe:: x == y

            Checks if two verification levels are equal.
        .. describe:: x != y

            Checks if two verification levels are not equal.
        .. describe:: x > y

            Checks if a verification level is higher than another.
        .. describe:: x < y

            Checks if a verification level is lower than another.
        .. describe:: x >= y

            Checks if a verification level is higher or equal to another.
        .. describe:: x <= y

            Checks if a verification level is lower or equal to another.
    """

    none = 0
    """No criteria set."""
    low = 1
    """Member must have a verified email on their Discord account."""
    medium = 2
    """Member must have a verified email and be registered on Discord for more than five minutes."""
    high = 3
    """Member must have a verified email, be registered on Discord for more
    than five minutes, and be a member of the guild itself for more than ten minutes.
    """
    highest = 4
    """Member must have a verified phone on their Discord account."""

    def __str__(self) -> str:
        return self.name


class ContentFilter(Enum, comparable=True):
    """Specifies a :class:`Guild`\\'s explicit content filter, which is the machine
    learning algorithms that Discord uses to detect if an image contains NSFW content.

    .. collapse:: operations

        .. describe:: x == y

            Checks if two content filter levels are equal.
        .. describe:: x != y

            Checks if two content filter levels are not equal.
        .. describe:: x > y

            Checks if a content filter level is higher than another.
        .. describe:: x < y

            Checks if a content filter level is lower than another.
        .. describe:: x >= y

            Checks if a content filter level is higher or equal to another.
        .. describe:: x <= y

            Checks if a content filter level is lower or equal to another.
    """

    disabled = 0
    """The guild does not have the content filter enabled."""
    no_role = 1
    """The guild has the content filter enabled for members without a role."""
    all_members = 2
    """The guild has the content filter enabled for every member."""

    def __str__(self) -> str:
        return self.name


class Status(Enum):
    """Specifies a :class:`Member`\\'s status."""

    online = "online"
    """The member is online."""
    offline = "offline"
    """The member is offline."""
    idle = "idle"
    """The member is idle."""
    dnd = "dnd"
    """The member is "Do Not Disturb"."""
    do_not_disturb = "dnd"
    """An alias for :attr:`dnd`."""
    invisible = "invisible"
    """The member is "invisible". In reality, this is only used in sending
    a presence a la :meth:`Client.change_presence`. When you receive a
    user's presence this will be :attr:`offline` instead.
    """
    streaming = "streaming"
    """The member is live streaming to Twitch or YouTube.

    .. versionadded:: 2.3
    """

    def __str__(self) -> str:
        return self.value


class DefaultAvatar(Enum):
    """Represents the default avatar of a Discord :class:`User`."""

    blurple = 0
    """Represents the default avatar with the color blurple. See also :attr:`Colour.blurple`."""
    grey = 1
    """Represents the default avatar with the color grey. See also :attr:`Colour.greyple`."""
    gray = 1
    """An alias for :attr:`grey`."""
    green = 2
    """Represents the default avatar with the color green. See also :attr:`Colour.green`."""
    orange = 3
    """Represents the default avatar with the color orange. See also :attr:`Colour.orange`."""
    red = 4
    """Represents the default avatar with the color red. See also :attr:`Colour.red`."""
    fuchsia = 5
    """Represents the default avatar with the color fuchsia. See also :attr:`Colour.fuchsia`.

    .. versionadded:: 2.9
    """

    def __str__(self) -> str:
        return self.name


class NotificationLevel(Enum, comparable=True):
    """Specifies whether a :class:`Guild` has notifications on for all messages or mentions only by default.

    .. collapse:: operations

        .. describe:: x == y

            Checks if two notification levels are equal.
        .. describe:: x != y

            Checks if two notification levels are not equal.
        .. describe:: x > y

            Checks if a notification level is higher than another.
        .. describe:: x < y

            Checks if a notification level is lower than another.
        .. describe:: x >= y

            Checks if a notification level is higher or equal to another.
        .. describe:: x <= y

            Checks if a notification level is lower or equal to another.
    """

    all_messages = 0
    """Members receive notifications for every message regardless of them being mentioned."""
    only_mentions = 1
    """Members receive notifications for messages they are mentioned in."""


class AuditLogActionCategory(Enum):
    """Represents the category that the :class:`AuditLogAction` belongs to.

    This can be retrieved via :attr:`AuditLogEntry.category`.
    """

    create = 1
    """The action is the creation of something."""
    delete = 2
    """The action is the deletion of something."""
    update = 3
    """The action is the update of something."""


# NOTE: these fields are only fully documented in audit_logs.rst,
# as the docstrings alone would be ~1000-1500 additional lines
class AuditLogAction(Enum):
    """Represents the type of action being done for a :class:`AuditLogEntry`\\,
    which is retrievable via :meth:`Guild.audit_logs` or via the :func:`on_audit_log_entry_create` event.
    """

    # fmt: off
    guild_update                          = 1
    channel_create                        = 10
    channel_update                        = 11
    channel_delete                        = 12
    overwrite_create                      = 13
    overwrite_update                      = 14
    overwrite_delete                      = 15
    kick                                  = 20
    member_prune                          = 21
    ban                                   = 22
    unban                                 = 23
    member_update                         = 24
    member_role_update                    = 25
    member_move                           = 26
    member_disconnect                     = 27
    bot_add                               = 28
    role_create                           = 30
    role_update                           = 31
    role_delete                           = 32
    invite_create                         = 40
    invite_update                         = 41
    invite_delete                         = 42
    webhook_create                        = 50
    webhook_update                        = 51
    webhook_delete                        = 52
    emoji_create                          = 60
    emoji_update                          = 61
    emoji_delete                          = 62
    message_delete                        = 72
    message_bulk_delete                   = 73
    message_pin                           = 74
    message_unpin                         = 75
    integration_create                    = 80
    integration_update                    = 81
    integration_delete                    = 82
    stage_instance_create                 = 83
    stage_instance_update                 = 84
    stage_instance_delete                 = 85
    sticker_create                        = 90
    sticker_update                        = 91
    sticker_delete                        = 92
    guild_scheduled_event_create          = 100
    guild_scheduled_event_update          = 101
    guild_scheduled_event_delete          = 102
    thread_create                         = 110
    thread_update                         = 111
    thread_delete                         = 112
    application_command_permission_update = 121
    soundboard_sound_create               = 130
    soundboard_sound_update               = 131
    soundboard_sound_delete               = 132
    automod_rule_create                   = 140
    automod_rule_update                   = 141
    automod_rule_delete                   = 142
    automod_block_message                 = 143
    automod_send_alert_message            = 144
    automod_timeout                       = 145
    creator_monetization_request_created  = 150
    creator_monetization_terms_accepted   = 151
    # fmt: on

    @property
    def category(self) -> Optional[AuditLogActionCategory]:
        # fmt: off
        lookup: Dict[AuditLogAction, Optional[AuditLogActionCategory]] = {
            AuditLogAction.guild_update:                          AuditLogActionCategory.update,
            AuditLogAction.channel_create:                        AuditLogActionCategory.create,
            AuditLogAction.channel_update:                        AuditLogActionCategory.update,
            AuditLogAction.channel_delete:                        AuditLogActionCategory.delete,
            AuditLogAction.overwrite_create:                      AuditLogActionCategory.create,
            AuditLogAction.overwrite_update:                      AuditLogActionCategory.update,
            AuditLogAction.overwrite_delete:                      AuditLogActionCategory.delete,
            AuditLogAction.kick:                                  None,
            AuditLogAction.member_prune:                          None,
            AuditLogAction.ban:                                   None,
            AuditLogAction.unban:                                 None,
            AuditLogAction.member_update:                         AuditLogActionCategory.update,
            AuditLogAction.member_role_update:                    AuditLogActionCategory.update,
            AuditLogAction.member_move:                           None,
            AuditLogAction.member_disconnect:                     None,
            AuditLogAction.bot_add:                               None,
            AuditLogAction.role_create:                           AuditLogActionCategory.create,
            AuditLogAction.role_update:                           AuditLogActionCategory.update,
            AuditLogAction.role_delete:                           AuditLogActionCategory.delete,
            AuditLogAction.invite_create:                         AuditLogActionCategory.create,
            AuditLogAction.invite_update:                         AuditLogActionCategory.update,
            AuditLogAction.invite_delete:                         AuditLogActionCategory.delete,
            AuditLogAction.webhook_create:                        AuditLogActionCategory.create,
            AuditLogAction.webhook_update:                        AuditLogActionCategory.update,
            AuditLogAction.webhook_delete:                        AuditLogActionCategory.delete,
            AuditLogAction.emoji_create:                          AuditLogActionCategory.create,
            AuditLogAction.emoji_update:                          AuditLogActionCategory.update,
            AuditLogAction.emoji_delete:                          AuditLogActionCategory.delete,
            AuditLogAction.message_delete:                        AuditLogActionCategory.delete,
            AuditLogAction.message_bulk_delete:                   AuditLogActionCategory.delete,
            AuditLogAction.message_pin:                           None,
            AuditLogAction.message_unpin:                         None,
            AuditLogAction.integration_create:                    AuditLogActionCategory.create,
            AuditLogAction.integration_update:                    AuditLogActionCategory.update,
            AuditLogAction.integration_delete:                    AuditLogActionCategory.delete,
            AuditLogAction.stage_instance_create:                 AuditLogActionCategory.create,
            AuditLogAction.stage_instance_update:                 AuditLogActionCategory.update,
            AuditLogAction.stage_instance_delete:                 AuditLogActionCategory.delete,
            AuditLogAction.sticker_create:                        AuditLogActionCategory.create,
            AuditLogAction.sticker_update:                        AuditLogActionCategory.update,
            AuditLogAction.sticker_delete:                        AuditLogActionCategory.delete,
            AuditLogAction.thread_create:                         AuditLogActionCategory.create,
            AuditLogAction.thread_update:                         AuditLogActionCategory.update,
            AuditLogAction.thread_delete:                         AuditLogActionCategory.delete,
            AuditLogAction.guild_scheduled_event_create:          AuditLogActionCategory.create,
            AuditLogAction.guild_scheduled_event_update:          AuditLogActionCategory.update,
            AuditLogAction.guild_scheduled_event_delete:          AuditLogActionCategory.delete,
            AuditLogAction.application_command_permission_update: AuditLogActionCategory.update,
            AuditLogAction.soundboard_sound_create:               AuditLogActionCategory.create,
            AuditLogAction.soundboard_sound_update:               AuditLogActionCategory.update,
            AuditLogAction.soundboard_sound_delete:               AuditLogActionCategory.delete,
            AuditLogAction.automod_rule_create:                   AuditLogActionCategory.create,
            AuditLogAction.automod_rule_update:                   AuditLogActionCategory.update,
            AuditLogAction.automod_rule_delete:                   AuditLogActionCategory.delete,
            AuditLogAction.automod_block_message:                 None,
            AuditLogAction.automod_send_alert_message:            None,
            AuditLogAction.automod_timeout:                       None,
            AuditLogAction.creator_monetization_request_created:  None,
            AuditLogAction.creator_monetization_terms_accepted:   None,
        }
        # fmt: on
        return lookup[self]

    @property
    def target_type(self) -> Optional[str]:
        v = self.value
        if v == -1:  # pyright: ignore[reportUnnecessaryComparison]
            return "all"
        elif v < 10:
            return "guild"
        elif v < 20:
            return "channel"
        elif v < 30:
            return "user"
        elif v < 40:
            return "role"
        elif v < 50:
            return "invite"
        elif v < 60:
            return "webhook"
        elif v < 70:
            return "emoji"
        elif v == 73:
            return "channel"
        elif v < 80:
            return "message"
        elif v < 83:
            return "integration"
        elif v < 90:
            return "stage_instance"
        elif v < 93:
            return "sticker"
        elif v < 103:
            return "guild_scheduled_event"
        elif v < 113:
            return "thread"
        elif v < 122:
            return "application_command_or_integration"
        elif v < 140:
            return None
        elif v < 143:
            return "automod_rule"
        elif v < 146:
            return "user"
        elif v < 152:
            return None
        else:
            return None


class UserFlags(Enum):
    """Represents Discord user flags."""

    staff = 1 << 0
    """The user is a Discord Employee."""
    partner = 1 << 1
    """The user is a Discord Partner."""
    hypesquad = 1 << 2
    """The user is a HypeSquad Events member."""
    bug_hunter = 1 << 3
    """The user is a Bug Hunter."""
    mfa_sms = 1 << 4
    """The user has SMS recovery for Multi Factor Authentication enabled."""
    premium_promo_dismissed = 1 << 5
    """The user has dismissed the Discord Nitro promotion."""
    hypesquad_bravery = 1 << 6
    """The user is a HypeSquad Bravery member."""
    hypesquad_brilliance = 1 << 7
    """The user is a HypeSquad Brilliance member."""
    hypesquad_balance = 1 << 8
    """The user is a HypeSquad Balance member."""
    early_supporter = 1 << 9
    """The user is an Early Supporter."""
    team_user = 1 << 10
    """The user is a Team User."""
    system = 1 << 12
    """The user is a system user (i.e. represents Discord officially)."""
    has_unread_urgent_messages = 1 << 13
    """The user has an unread system message."""
    bug_hunter_level_2 = 1 << 14
    """The user is a Bug Hunter Level 2."""
    verified_bot = 1 << 16
    """The user is a Verified Bot."""
    verified_bot_developer = 1 << 17
    """The user is an Early Verified Bot Developer."""
    discord_certified_moderator = 1 << 18
    """The user is a Discord Certified Moderator."""
    http_interactions_bot = 1 << 19
    """The user is a bot that only uses HTTP interactions.

    .. versionadded:: 2.3
    """
    spammer = 1 << 20
    """The user is marked as a spammer.

    .. versionadded:: 2.3
    """
    active_developer = 1 << 22
    """The user is an Active Developer.

    .. versionadded:: 2.8
    """


class ActivityType(Enum):
    """Specifies the type of :class:`Activity`. This is used to check how to
    interpret the activity itself.
    """

    unknown = -1
    """An unknown activity type. This should generally not happen."""
    playing = 0
    """A "Playing" activity type."""
    streaming = 1
    """A "Streaming" activity type."""
    listening = 2
    """A "Listening" activity type."""
    watching = 3
    """A "Watching" activity type."""
    custom = 4
    """A custom activity type."""
    competing = 5
    """A competing activity type.

    .. versionadded:: 1.5
    """

    def __int__(self) -> int:
        return self.value


class TeamMembershipState(Enum):
    """Represents the membership state of a team member retrieved through :func:`Client.application_info`.

    .. versionadded:: 1.3
    """

    invited = 1
    """Represents an invited member."""
    accepted = 2
    """Represents a member currently in the team."""


class TeamMemberRole(Enum):
    """Represents the role of a team member retrieved through :func:`Client.application_info`.

    .. versionadded:: 2.10
    """

    admin = "admin"
    """Admins have the most permissions. An admin can only take destructive actions
    on the team or team-owned apps if they are the team owner.
    """
    developer = "developer"
    """Developers can access information about a team and team-owned applications,
    and take limited actions on them, like configuring interaction
    endpoints or resetting the bot token.
    """
    read_only = "read_only"
    """Read-only members can access information about a team and team-owned applications."""

    def __str__(self) -> str:
        return self.name


class WebhookType(Enum):
    """Represents the type of webhook that can be received.

    .. versionadded:: 1.3
    """

    incoming = 1
    """Represents a webhook that can post messages to channels with a token."""
    channel_follower = 2
    """Represents a webhook that is internally managed by Discord, used for following channels."""
    application = 3
    """Represents a webhook that is used for interactions or applications.

    .. versionadded:: 2.0
    """


class ExpireBehaviour(Enum):
    """Represents the behaviour the :class:`Integration` should perform
    when a user's subscription has finished.

    There is an alias for this called ``ExpireBehavior``.

    .. versionadded:: 1.4
    """

    remove_role = 0
    """This will remove the :attr:`StreamIntegration.role` from the user
    when their subscription is finished.
    """
    kick = 1
    """This will kick the user when their subscription is finished."""


ExpireBehavior = ExpireBehaviour


class StickerType(Enum):
    """Represents the type of sticker.

    .. versionadded:: 2.0
    """

    standard = 1
    """Represents a standard sticker that all users can use."""
    guild = 2
    """Represents a custom sticker created in a guild."""


class StickerFormatType(Enum):
    """Represents the type of sticker images.

    .. versionadded:: 1.6
    """

    png = 1
    """Represents a sticker with a png image."""
    apng = 2
    """Represents a sticker with an apng image."""
    lottie = 3
    """Represents a sticker with a lottie image."""
    gif = 4
    """Represents a sticker with a gif image.

    .. versionadded:: 2.8
    """

    @property
    def file_extension(self) -> str:
        return STICKER_FORMAT_LOOKUP[self]


STICKER_FORMAT_LOOKUP: Dict[StickerFormatType, str] = {
    StickerFormatType.png: "png",
    StickerFormatType.apng: "png",
    StickerFormatType.lottie: "json",
    StickerFormatType.gif: "gif",
}


class InviteType(Enum):
    """Represents the type of an invite.

    .. versionadded:: 2.10
    """

    guild = 0
    """Represents an invite to a guild."""
    group_dm = 1
    """Represents an invite to a group channel."""
    friend = 2
    """Represents a friend invite."""


class InviteTarget(Enum):
    """Represents the invite type for voice channel invites.

    .. versionadded:: 2.0
    """

    unknown = 0
    """The invite doesn't target anyone or anything."""
    stream = 1
    """A stream invite that targets a user."""
    embedded_application = 2
    """A stream invite that targets an embedded application."""


class InteractionType(Enum):
    """Specifies the type of :class:`Interaction`.

    .. versionadded:: 2.0
    """

    ping = 1
    """Represents Discord pinging to see if the interaction response server is alive."""
    application_command = 2
    """Represents an application command interaction."""
    component = 3
    """Represents a component based interaction, i.e. using the Discord Bot UI Kit."""
    application_command_autocomplete = 4
    """Represents an application command autocomplete interaction."""
    modal_submit = 5
    """Represents a modal submit interaction."""


class InteractionResponseType(Enum):
    """Specifies the response type for the interaction.

    .. versionadded:: 2.0
    """

    pong = 1
    """Pongs the interaction when given a ping.

    See also :meth:`InteractionResponse.pong`.
    """
    channel_message = 4
    """Responds to the interaction with a message.

    See also :meth:`InteractionResponse.send_message`.
    """
    deferred_channel_message = 5
    """Responds to the interaction with a message at a later time.

    See also :meth:`InteractionResponse.defer`.
    """
    deferred_message_update = 6
    """Acknowledges the component interaction with a promise that
    the message will update later (though there is no need to actually update the message).

    See also :meth:`InteractionResponse.defer`.
    """
    message_update = 7
    """Responds to the interaction by editing the message.

    See also :meth:`InteractionResponse.edit_message`.
    """
    application_command_autocomplete_result = 8
    """Responds to the autocomplete interaction with suggested choices.

    See also :meth:`InteractionResponse.autocomplete`.
    """
    modal = 9
    """Responds to the interaction by displaying a modal.

    See also :meth:`InteractionResponse.send_modal`.

    .. versionadded:: 2.4
    """
    premium_required = 10
    """Responds to the interaction with a message containing an upgrade button.
    Only available for applications with monetization enabled.

    See also :meth:`InteractionResponse.require_premium`.

    .. versionadded:: 2.10
    """


class VideoQualityMode(Enum):
    """Represents the camera video quality mode for voice channel participants.

    .. versionadded:: 2.0
    """

    auto = 1
    """Represents auto camera video quality."""
    full = 2
    """Represents full camera video quality."""

    def __int__(self) -> int:
        return self.value


class ComponentType(Enum):
    """Represents the type of component.

    .. versionadded:: 2.0
    """

    action_row = 1
    """Represents the group component which holds different components in a row."""
    button = 2
    """Represents a button component."""
    string_select = 3
    """Represents a string select component.

    .. versionadded:: 2.7
    """
    select = 3  # backwards compatibility
    """An alias of :attr:`string_select`."""
    text_input = 4
    """Represents a text input component."""
    user_select = 5
    """Represents a user select component.

    .. versionadded:: 2.7
    """
    role_select = 6
    """Represents a role select component.

    .. versionadded:: 2.7
    """
    mentionable_select = 7
    """Represents a mentionable (user/member/role) select component.

    .. versionadded:: 2.7
    """
    channel_select = 8
    """Represents a channel select component.

    .. versionadded:: 2.7
    """

    def __int__(self) -> int:
        return self.value


class ButtonStyle(Enum):
    """Represents the style of the button component.

    .. versionadded:: 2.0
    """

    primary = 1
    """Represents a blurple button for the primary action."""
    secondary = 2
    """Represents a grey button for the secondary action."""
    success = 3
    """Represents a green button for a successful action."""
    danger = 4
    """Represents a red button for a dangerous action."""
    link = 5
    """Represents a link button."""

    # Aliases
    blurple = 1
    """An alias for :attr:`primary`."""
    grey = 2
    """An alias for :attr:`secondary`."""
    gray = 2
    """An alias for :attr:`secondary`."""
    green = 3
    """An alias for :attr:`success`."""
    red = 4
    """An alias for :attr:`danger`."""
    url = 5
    """An alias for :attr:`link`."""

    def __int__(self) -> int:
        return self.value


class TextInputStyle(Enum):
    """Represents a style of the text input component.

    .. versionadded:: 2.4
    """

    short = 1
    """Represents a single-line text input component."""
    paragraph = 2
    """Represents a multi-line text input component."""

    # Aliases
    single_line = 1
    """An alias for :attr:`short`."""
    multi_line = 2
    """An alias for :attr:`paragraph`."""
    long = 2
    """An alias for :attr:`paragraph`."""

    def __int__(self) -> int:
        return self.value


class SelectDefaultValueType(Enum):
    """Represents the type of a :class:`SelectDefaultValue`.

    .. versionadded:: 2.10
    """

    user = "user"
    """Represents a user/member."""
    role = "role"
    """Represents a role."""
    channel = "channel"
    """Represents a channel."""

    def __str__(self) -> str:
        return self.value


class ApplicationCommandType(Enum):
    """Represents the type of an application command.

    .. versionadded:: 2.1
    """

    chat_input = 1
    """Represents a slash command."""
    user = 2
    """Represents a user command from the context menu."""
    message = 3
    """Represents a message command from the context menu."""


class ApplicationCommandPermissionType(Enum):
    """Represents the type of a permission of an application command.

    .. versionadded:: 2.5
    """

    role = 1
    """Represents a permission that affects roles."""
    user = 2
    """Represents a permission that affects users."""
    channel = 3
    """Represents a permission that affects channels."""

    def __int__(self) -> int:
        return self.value


class OptionType(Enum):
    """Represents the type of an option.

    .. versionadded:: 2.1
    """

    sub_command = 1
    """Represents a sub command of the main command or group."""
    sub_command_group = 2
    """Represents a sub command group of the main command."""
    string = 3
    """Represents a string option."""
    integer = 4
    """Represents an integer option."""
    boolean = 5
    """Represents a boolean option."""
    user = 6
    """Represents a user option."""
    channel = 7
    """Represents a channel option."""
    role = 8
    """Represents a role option."""
    mentionable = 9
    """Represents a role + user option."""
    number = 10
    """Represents a float option."""
    attachment = 11
    """Represents an attachment option.

    .. versionadded:: 2.4
    """


class StagePrivacyLevel(Enum):
    """Represents a stage instance's privacy level.

    .. versionadded:: 2.0
    """

    public = 1
    """The stage instance can be joined by external users.

    .. deprecated:: 2.5
        Public stages are no longer supported by discord.
    """
    closed = 2
    """The stage instance can only be joined by members of the guild."""
    guild_only = 2
    """Alias for :attr:`.closed`"""


class NSFWLevel(Enum, comparable=True):
    """Represents the NSFW level of a guild.

    .. versionadded:: 2.0

    .. collapse:: operations

        .. describe:: x == y

            Checks if two NSFW levels are equal.
        .. describe:: x != y

            Checks if two NSFW levels are not equal.
        .. describe:: x > y

            Checks if an NSFW level is higher than another.
        .. describe:: x < y

            Checks if an NSFW level is lower than another.
        .. describe:: x >= y

            Checks if an NSFW level is higher or equal to another.
        .. describe:: x <= y

            Checks if an NSFW level is lower or equal to another.
    """

    default = 0
    """The guild has not been categorised yet."""
    explicit = 1
    """The guild contains NSFW content."""
    safe = 2
    """The guild does not contain any NSFW content."""
    age_restricted = 3
    """The guild may contain NSFW content."""


class GuildScheduledEventEntityType(Enum):
    """Represents the type of a guild scheduled event entity.

    .. versionadded:: 2.3
    """

    stage_instance = 1
    """The guild scheduled event will take place in a stage channel."""
    voice = 2
    """The guild scheduled event will take place in a voice channel."""
    external = 3
    """The guild scheduled event will take place in a custom location."""


class GuildScheduledEventStatus(Enum):
    """Represents the status of a guild scheduled event.

    .. versionadded:: 2.3
    """

    scheduled = 1
    """Represents a scheduled event."""
    active = 2
    """Represents an active event."""
    completed = 3
    """Represents a completed event."""
    canceled = 4
    """Represents a canceled event."""
    cancelled = 4
    """An alias for :attr:`canceled`.

    .. versionadded:: 2.6
    """


class GuildScheduledEventPrivacyLevel(Enum):
    """Represents the privacy level of a guild scheduled event.

    .. versionadded:: 2.3
    """

    guild_only = 2
    """The guild scheduled event is only for a specific guild."""


class ThreadArchiveDuration(Enum):
    """Represents the automatic archive duration of a thread in minutes.

    .. versionadded:: 2.3
    """

    hour = 60
    """The thread will archive after an hour of inactivity."""
    day = 1440
    """The thread will archive after a day of inactivity."""
    three_days = 4320
    """The thread will archive after three days of inactivity."""
    week = 10080
    """The thread will archive after a week of inactivity."""

    def __int__(self) -> int:
        return self.value


class WidgetStyle(Enum):
    """Represents the supported widget image styles.

    .. versionadded:: 2.5
    """

    shield = "shield"
    """A shield style image with a Discord icon and the online member count."""
    banner1 = "banner1"
    """A large image with guild icon, name and online member count and a footer."""
    banner2 = "banner2"
    """A small image with guild icon, name and online member count."""
    banner3 = "banner3"
    """A large image with guild icon, name and online member count and a footer,
    with a "Chat Now" label on the right.
    """
    banner4 = "banner4"
    """A large image with a large Discord logo, guild icon, name and online member count,
    with a "Join My Server" label at the bottom.
    """

    def __str__(self) -> str:
        return self.value


# reference: https://discord.com/developers/docs/reference#locales
class Locale(Enum):
    """Represents supported locales by Discord.

    .. versionadded:: 2.5
    """

    bg = "bg"
    """The ``bg`` (Bulgarian) locale."""
    cs = "cs"
    """The ``cs`` (Czech) locale."""
    da = "da"
    """The ``da`` (Danish) locale."""
    de = "de"
    """The ``de`` (German) locale."""
    el = "el"
    """The ``el`` (Greek) locale."""
    en_GB = "en-GB"
    """The ``en-GB`` (English, UK) locale."""
    en_US = "en-US"
    """The ``en-US`` (English, US) locale."""
    es_ES = "es-ES"
    """The ``es-ES`` (Spanish) locale."""
    es_LATAM = "es-419"
    """The ``es-419`` (Spanish, LATAM) locale.

    .. versionadded:: 2.10
    """
    fi = "fi"
    """The ``fi`` (Finnish) locale."""
    fr = "fr"
    """The ``fr`` (French) locale."""
    hi = "hi"
    """The ``hi`` (Hindi) locale."""
    hr = "hr"
    """The ``hr`` (Croatian) locale."""
    hu = "hu"
    """The ``hu`` (Hungarian) locale."""
    id = "id"
    """The ``id`` (Indonesian) locale.

    .. versionadded:: 2.8
    """
    it = "it"
    """The ``it`` (Italian) locale."""
    ja = "ja"
    """The ``ja`` (Japanese) locale."""
    ko = "ko"
    """The ``ko`` (Korean) locale."""
    lt = "lt"
    """The ``lt`` (Lithuanian) locale."""
    nl = "nl"
    """The ``nl`` (Dutch) locale."""
    no = "no"
    """The ``no`` (Norwegian) locale."""
    pl = "pl"
    """The ``pl`` (Polish) locale."""
    pt_BR = "pt-BR"
    """The ``pt-BR`` (Portuguese) locale."""
    ro = "ro"
    """The ``ro`` (Romanian) locale."""
    ru = "ru"
    """The ``ru`` (Russian) locale."""
    sv_SE = "sv-SE"
    """The ``sv-SE`` (Swedish) locale."""
    th = "th"
    """The ``th`` (Thai) locale."""
    tr = "tr"
    """The ``tr`` (Turkish) locale."""
    uk = "uk"
    """The ``uk`` (Ukrainian) locale."""
    vi = "vi"
    """The ``vi`` (Vietnamese) locale."""
    zh_CN = "zh-CN"
    """The ``zh-CN`` (Chinese, China) locale."""
    zh_TW = "zh-TW"
    """The ``zh-TW`` (Chinese, Taiwan) locale."""

    def __str__(self) -> str:
        return self.value


class AutoModActionType(Enum):
    """Represents the type of action an auto moderation rule will take upon execution.

    .. versionadded:: 2.6
    """

    block_message = 1
    """The rule will prevent matching messages from being posted."""
    send_alert_message = 2
    """The rule will send an alert to a specified channel."""
    timeout = 3
    """The rule will timeout the user that sent the message.

    .. note::
        This action type is only available for rules with trigger type
        :attr:`~AutoModTriggerType.keyword` or :attr:`~AutoModTriggerType.mention_spam`,
        and :attr:`~Permissions.moderate_members` permissions are required to use it.
    """


class AutoModEventType(Enum):
    """Represents the type of event/context an auto moderation rule will be checked in.

    .. versionadded:: 2.6
    """

    message_send = 1
    """The rule will apply when a member sends or edits a message in the guild."""


class AutoModTriggerType(Enum):
    """Represents the type of content that can trigger an auto moderation rule.

    .. versionadded:: 2.6

    .. versionchanged:: 2.9
        Removed obsolete ``harmful_link`` type.
    """

    keyword = 1
    """The rule will filter messages based on a custom keyword list.

    This trigger type requires additional :class:`metadata <AutoModTriggerMetadata>`.
    """

    if not TYPE_CHECKING:
        harmful_link = 2  # obsolete/deprecated

    spam = 3
    """The rule will filter messages suspected of being spam."""
    keyword_preset = 4
    """The rule will filter messages based on predefined lists containing commonly flagged words.

    This trigger type requires additional :class:`metadata <AutoModTriggerMetadata>`.
    """
    mention_spam = 5
    """The rule will filter messages based on the number of member/role mentions they contain.

    This trigger type requires additional :class:`metadata <AutoModTriggerMetadata>`.
    """


class ThreadSortOrder(Enum):
    """Represents the sort order of threads in a :class:`ForumChannel` or :class:`MediaChannel`.

    .. versionadded:: 2.6
    """

    latest_activity = 0
    """Sort forum threads by activity."""
    creation_date = 1
    """Sort forum threads by creation date/time (from newest to oldest)."""


class ThreadLayout(Enum):
    """Represents the layout of threads in :class:`ForumChannel`\\s.

    .. versionadded:: 2.8
    """

    not_set = 0
    """No preferred layout has been set."""
    list_view = 1
    """Display forum threads in a text-focused list."""
    gallery_view = 2
    """Display forum threads in a media-focused collection of tiles."""


class Event(Enum):
    """
    Represents all the events of the library.

    These offer to register listeners/events in a more pythonic way; additionally autocompletion and documentation are both supported.

    .. versionadded:: 2.8
    """

    connect = "connect"
    """Called when the client has successfully connected to Discord.
    Represents the :func:`on_connect` event.
    """
    disconnect = "disconnect"
    """Called when the client has disconnected from Discord, or a connection attempt to Discord has failed.
    Represents the :func:`on_disconnect` event.
    """
    error = "error"
    """Called when an uncaught exception occurred.
    Represents the :func:`on_error` event.
    """
    gateway_error = "gateway_error"
    """Called when a known gateway event cannot be parsed.
    Represents the :func:`on_gateway_error` event.
    """
    ready = "ready"
    """Called when the client is done preparing the data received from Discord.
    Represents the :func:`on_ready` event.
    """
    resumed = "resumed"
    """Called when the client has resumed a session.
    Represents the :func:`on_resumed` event.
    """
    shard_connect = "shard_connect"
    """Called when a shard has successfully connected to Discord.
    Represents the :func:`on_shard_connect` event.
    """
    shard_disconnect = "shard_disconnect"
    """Called when a shard has disconnected from Discord.
    Represents the :func:`on_shard_disconnect` event.
    """
    shard_ready = "shard_ready"
    """Called when a shard has become ready.
    Represents the :func:`on_shard_ready` event.
    """
    shard_resumed = "shard_resumed"
    """Called when a shard has resumed a session.
    Represents the :func:`on_shard_resumed` event.
    """
    socket_event_type = "socket_event_type"
    """Called whenever a websocket event is received from the WebSocket.
    Represents the :func:`on_socket_event_type` event.
    """
    socket_raw_receive = "socket_raw_receive"
    """Called whenever a message is completely received from the WebSocket, before it's processed and parsed.
    Represents the :func:`on_socket_raw_receive` event.
    """
    socket_raw_send = "socket_raw_send"
    """Called whenever a send operation is done on the WebSocket before the message is sent.
    Represents the :func:`on_socket_raw_send` event.
    """
    guild_channel_create = "guild_channel_create"
    """Called whenever a guild channel is created.
    Represents the :func:`on_guild_channel_create` event.
    """
    guild_channel_update = "guild_channel_update"
    """Called whenever a guild channel is updated.
    Represents the :func:`on_guild_channel_update` event.
    """
    guild_channel_delete = "guild_channel_delete"
    """Called whenever a guild channel is deleted.
    Represents the :func:`on_guild_channel_delete` event.
    """
    guild_channel_pins_update = "guild_channel_pins_update"
    """Called whenever a message is pinned or unpinned from a guild channel.
    Represents the :func:`on_guild_channel_pins_update` event.
    """
    invite_create = "invite_create"
    """Called when an :class:`Invite` is created.
    Represents the :func:`.on_invite_create` event.
    """
    invite_delete = "invite_delete"
    """Called when an Invite is deleted.
    Represents the :func:`.on_invite_delete` event.
    """
    private_channel_update = "private_channel_update"
    """Called whenever a private group DM is updated.
    Represents the :func:`on_private_channel_update` event.
    """
    private_channel_pins_update = "private_channel_pins_update"
    """Called whenever a message is pinned or unpinned from a private channel.
    Represents the :func:`on_private_channel_pins_update` event.
    """
    webhooks_update = "webhooks_update"
    """Called whenever a webhook is created, modified, or removed from a guild channel.
    Represents the :func:`on_webhooks_update` event.
    """
    thread_create = "thread_create"
    """Called whenever a thread is created.
    Represents the :func:`on_thread_create` event.
    """
    thread_update = "thread_update"
    """Called when a thread is updated.
    Represents the :func:`on_thread_update` event.
    """
    thread_delete = "thread_delete"
    """Called when a thread is deleted.
    Represents the :func:`on_thread_delete` event.
    """
    thread_join = "thread_join"
    """Called whenever the bot joins a thread or gets access to a thread.
    Represents the :func:`on_thread_join` event.
    """
    thread_remove = "thread_remove"
    """Called whenever a thread is removed. This is different from a thread being deleted.
    Represents the :func:`on_thread_remove` event.
    """
    thread_member_join = "thread_member_join"
    """Called when a `ThreadMember` joins a `Thread`.
    Represents the :func:`on_thread_member_join` event.
    """
    thread_member_remove = "thread_member_remove"
    """Called when a `ThreadMember` leaves a `Thread`.
    Represents the :func:`on_thread_member_remove` event.
    """
    raw_thread_member_remove = "raw_thread_member_remove"
    """Called when a `ThreadMember` leaves `Thread` regardless of the thread member cache.
    Represents the :func:`on_raw_thread_member_remove` event.
    """
    raw_thread_update = "raw_thread_update"
    """Called whenever a thread is updated regardless of the state of the internal thread cache.
    Represents the :func:`on_raw_thread_update` event.
    """
    raw_thread_delete = "raw_thread_delete"
    """Called whenever a thread is deleted regardless of the state of the internal thread cache.
    Represents the :func:`on_raw_thread_delete` event.
    """
    guild_join = "guild_join"
    """Called when a `Guild` is either created by the `Client` or when the Client joins a guild.
    Represents the :func:`on_guild_join` event.
    """
    guild_remove = "guild_remove"
    """Called when a `Guild` is removed from the :class:`Client`.
    Represents the :func:`on_guild_remove` event.
    """
    guild_update = "guild_update"
    """Called when a `Guild` updates.
    Represents the :func:`on_guild_update` event.
    """
    guild_available = "guild_available"
    """Called when a guild becomes available.
    Represents the :func:`on_guild_available` event.
    """
    guild_unavailable = "guild_unavailable"
    """Called when a guild becomes unavailable.
    Represents the :func:`on_guild_unavailable` event.
    """
    guild_role_create = "guild_role_create"
    """Called when a `Guild` creates a new `Role`.
    Represents the :func:`on_guild_role_create` event.
    """
    guild_role_delete = "guild_role_delete"
    """Called when a `Guild` deletes a `Role`.
    Represents the :func:`on_guild_role_delete` event.
    """
    guild_role_update = "guild_role_update"
    """Called when a `Guild` updates a `Role`.
    Represents the :func:`on_guild_role_update` event.
    """
    guild_emojis_update = "guild_emojis_update"
    """Called when a `Guild` adds or removes `Emoji`.
    Represents the :func:`on_guild_emojis_update` event.
    """
    guild_stickers_update = "guild_stickers_update"
    """Called when a `Guild` updates its stickers.
    Represents the :func:`on_guild_stickers_update` event.
    """
    guild_soundboard_sounds_update = "guild_soundboard_sounds_update"
    """Called when a `Guild` updates its soundboard sounds.
    Represents the :func:`on_guild_soundboard_sounds_update` event.

    .. versionadded:: 2.10
    """
    guild_integrations_update = "guild_integrations_update"
    """Called whenever an integration is created, modified, or removed from a guild.
    Represents the :func:`on_guild_integrations_update` event.
    """
    guild_scheduled_event_create = "guild_scheduled_event_create"
    """Called when a guild scheduled event is created.
    Represents the :func:`on_guild_scheduled_event_create` event.
    """
    guild_scheduled_event_update = "guild_scheduled_event_update"
    """Called when a guild scheduled event is updated.
    Represents the :func:`on_guild_scheduled_event_update` event.
    """
    guild_scheduled_event_delete = "guild_scheduled_event_delete"
    """Called when a guild scheduled event is deleted.
    Represents the :func:`on_guild_scheduled_event_delete` event.
    """
    guild_scheduled_event_subscribe = "guild_scheduled_event_subscribe"
    """Called when a user subscribes from a guild scheduled event.
    Represents the :func:`on_guild_scheduled_event_subscribe` event.
    """
    guild_scheduled_event_unsubscribe = "guild_scheduled_event_unsubscribe"
    """Called when a user unsubscribes from a guild scheduled event.
    Represents the :func:`on_guild_scheduled_event_unsubscribe` event.
    """
    raw_guild_scheduled_event_subscribe = "raw_guild_scheduled_event_subscribe"
    """Called when a user subscribes from a guild scheduled event regardless of the guild scheduled event cache.
    Represents the :func:`on_raw_guild_scheduled_event_subscribe` event.
    """
    raw_guild_scheduled_event_unsubscribe = "raw_guild_scheduled_event_unsubscribe"
    """Called when a user subscribes to or unsubscribes from a guild scheduled event regardless of the guild scheduled event cache.
    Represents the :func:`on_raw_guild_scheduled_event_unsubscribe` event.
    """
    application_command_permissions_update = "application_command_permissions_update"
    """Called when the permissions of an application command or the application-wide command permissions are updated.
    Represents the :func:`on_application_command_permissions_update` event.
    """
    automod_action_execution = "automod_action_execution"
    """Called when an auto moderation action is executed due to a rule triggering for a particular event.
    Represents the :func:`on_automod_action_execution` event.
    """
    automod_rule_create = "automod_rule_create"
    """Called when an `AutoModRule` is created.
    Represents the :func:`on_automod_rule_create` event.
    """
    automod_rule_update = "automod_rule_update"
    """Called when an `AutoModRule` is updated.
    Represents the :func:`on_automod_rule_update` event.
    """
    automod_rule_delete = "automod_rule_delete"
    """Called when an `AutoModRule` is deleted.
    Represents the :func:`on_automod_rule_delete` event.
    """
    audit_log_entry_create = "audit_log_entry_create"
    """Called when an audit log entry is created.
    Represents the :func:`on_audit_log_entry_create` event.
    """
    integration_create = "integration_create"
    """Called when an integration is created.
    Represents the :func:`on_integration_create` event.
    """
    integration_update = "integration_update"
    """Called when an integration is updated.
    Represents the :func:`on_integration_update` event.
    """
    raw_integration_delete = "raw_integration_delete"
    """Called when an integration is deleted.
    Represents the :func:`on_raw_integration_delete` event.
    """
    member_join = "member_join"
    """Called when a `Member` joins a `Guild`.
    Represents the :func:`on_member_join` event.
    """
    member_remove = "member_remove"
    """Called when a `Member` leaves a `Guild`.
    Represents the :func:`on_member_remove` event.
    """
    member_update = "member_update"
    """Called when a `Member` is updated in a `Guild`.
    Represents the :func:`on_member_update` event.
    """
    raw_member_remove = "raw_member_remove"
    """Called when a member leaves a `Guild` regardless of the member cache.
    Represents the :func:`on_raw_member_remove` event.
    """
    raw_member_update = "raw_member_update"
    """Called when a `Member` is updated in a `Guild` regardless of the member cache.
    Represents the :func:`on_raw_member_update` event.
    """
    member_ban = "member_ban"
    """Called when user gets banned from a `Guild`.
    Represents the :func:`on_member_ban` event.
    """
    member_unban = "member_unban"
    """Called when a `User` gets unbanned from a `Guild`.
    Represents the :func:`on_member_unban` event.
    """
    presence_update = "presence_update"
    """Called when a `Member` updates their presence.
    Represents the :func:`on_presence_update` event.
    """
    user_update = "user_update"
    """Called when a `User` is updated.
    Represents the :func:`on_user_update` event.
    """
    voice_state_update = "voice_state_update"
    """Called when a `Member` changes their `VoiceState`.
    Represents the :func:`on_voice_state_update` event.
    """
    voice_channel_effect = "voice_channel_effect"
    """Called when a `Member` sends an effect in a voice channel the bot is connected to.
    Represents the :func:`on_voice_channel_effect` event.

    .. versionadded:: 2.10
    """
    raw_voice_channel_effect = "raw_voice_channel_effect"
    """Called when a `Member` sends an effect in a voice channel the bot is connected to,
    regardless of the member cache.
    Represents the :func:`on_raw_voice_channel_effect` event.

    .. versionadded:: 2.10
    """
    stage_instance_create = "stage_instance_create"
    """Called when a `StageInstance` is created for a `StageChannel`.
    Represents the :func:`on_stage_instance_create` event.
    """
    stage_instance_delete = "stage_instance_delete"
    """Called when a `StageInstance` is deleted for a `StageChannel`.
    Represents the :func:`on_stage_instance_delete` event.
    """
    stage_instance_update = "stage_instance_update"
    """Called when a `StageInstance` is updated.
    Represents the :func:`on_stage_instance_update` event.
    """
    application_command = "application_command"
    """Called when an application command is invoked.
    Represents the :func:`on_application_command` event.
    """
    application_command_autocomplete = "application_command_autocomplete"
    """Called when an application command autocomplete is called.
    Represents the :func:`on_application_command_autocomplete` event.
    """
    button_click = "button_click"
    """Called when a button is clicked.
    Represents the :func:`on_button_click` event.
    """
    dropdown = "dropdown"
    """Called when a select menu is clicked.
    Represents the :func:`on_dropdown` event.
    """
    interaction = "interaction"
    """Called when an interaction happened.
    Represents the :func:`on_interaction` event.
    """
    message_interaction = "message_interaction"
    """Called when a message interaction happened.
    Represents the :func:`on_message_interaction` event.
    """
    modal_submit = "modal_submit"
    """Called when a modal is submitted.
    Represents the :func:`on_modal_submit` event.
    """
    message = "message"
    """Called when a `Message` is created and sent.
    Represents the :func:`on_message` event.
    """
    message_edit = "message_edit"
    """Called when a `Message` receives an update event.
    Represents the :func:`on_message_edit` event.
    """
    message_delete = "message_delete"
    """Called when a message is deleted.
    Represents the :func:`on_message_delete` event.
    """
    bulk_message_delete = "bulk_message_delete"
    """Called when messages are bulk deleted.
    Represents the :func:`on_bulk_message_delete` event.
    """
    poll_vote_add = "poll_vote_add"
    """Called when a vote is added on a `Poll`.
    Represents the :func:`on_poll_vote_add` event.
    """
    poll_vote_remove = "poll_vote_remove"
    """Called when a vote is removed from a `Poll`.
    Represents the :func:`on_poll_vote_remove` event.
    """
    raw_message_edit = "raw_message_edit"
    """Called when a message is edited regardless of the state of the internal message cache.
    Represents the :func:`on_raw_message_edit` event.
    """
    raw_message_delete = "raw_message_delete"
    """Called when a message is deleted regardless of the message being in the internal message cache or not.
    Represents the :func:`on_raw_message_delete` event.
    """
    raw_bulk_message_delete = "raw_bulk_message_delete"
    """Called when a bulk delete is triggered regardless of the messages being in the internal message cache or not.
    Represents the :func:`on_raw_bulk_message_delete` event.
    """
    raw_poll_vote_add = "raw_poll_vote_add"
    """Called when a vote is added on a `Poll` regardless of the internal message cache.
    Represents the :func:`on_raw_poll_vote_add` event.
    """
    raw_poll_vote_remove = "raw_poll_vote_remove"
    """Called when a vote is removed from a `Poll` regardless of the internal message cache.
    Represents the :func:`on_raw_poll_vote_remove` event.
    """
    reaction_add = "reaction_add"
    """Called when a message has a reaction added to it.
    Represents the :func:`on_reaction_add` event.
    """
    reaction_remove = "reaction_remove"
    """Called when a message has a reaction removed from it.
    Represents the :func:`on_reaction_remove` event.
    """
    reaction_clear = "reaction_clear"
    """Called when a message has all its reactions removed from it.
    Represents the :func:`on_reaction_clear` event.
    """
    reaction_clear_emoji = "reaction_clear_emoji"
    """Called when a message has a specific reaction removed from it.
    Represents the :func:`on_reaction_clear_emoji` event.
    """
    raw_presence_update = "raw_presence_update"
    """Called when a user's presence changes regardless of the state of the internal member cache.
    Represents the :func:`on_raw_presence_update` event.
    """
    raw_reaction_add = "raw_reaction_add"
    """Called when a message has a reaction added regardless of the state of the internal message cache.
    Represents the :func:`on_raw_reaction_add` event.
    """
    raw_reaction_remove = "raw_reaction_remove"
    """Called when a message has a reaction removed regardless of the state of the internal message cache.
    Represents the :func:`on_raw_reaction_remove` event.
    """
    raw_reaction_clear = "raw_reaction_clear"
    """Called when a message has all its reactions removed regardless of the state of the internal message cache.
    Represents the :func:`on_raw_reaction_clear` event.
    """
    raw_reaction_clear_emoji = "raw_reaction_clear_emoji"
    """Called when a message has a specific reaction removed from it regardless of the state of the internal message cache.
    Represents the :func:`on_raw_reaction_clear_emoji` event.
    """
    typing = "typing"
    """Called when someone begins typing a message.
    Represents the :func:`on_typing` event.
    """
    raw_typing = "raw_typing"
    """Called when someone begins typing a message regardless of whether `Intents.members` and `Intents.guilds` are enabled.
    Represents the :func:`on_raw_typing` event.
    """
    entitlement_create = "entitlement_create"
    """Called when a user subscribes to an SKU, creating a new :class:`Entitlement`.
    Represents the :func:`on_entitlement_create` event.

    .. versionadded:: 2.10
    """
    entitlement_update = "entitlement_update"
    """Called when a user's subscription renews.
    Represents the :func:`on_entitlement_update` event.

    .. versionadded:: 2.10
    """
    entitlement_delete = "entitlement_delete"
    """Called when a user's entitlement is deleted.
    Represents the :func:`on_entitlement_delete` event."""
    subscription_create = "subscription_create"
    """Called when a subscription for a premium app is created.
    Represents the :func:`on_subscription_create` event.

    .. versionadded:: 2.10
    """
    subscription_update = "subscription_update"
    """Called when a subscription for a premium app is updated.
    Represents the :func:`on_subscription_update` event.

    .. versionadded:: 2.10
    """
    subscription_delete = "subscription_delete"
    """Called when a subscription for a premium app is deleted.
    Represents the :func:`on_subscription_delete` event.

    .. versionadded:: 2.10
    """
    # ext.commands events
    command = "command"
    """Called when a command is found and is about to be invoked.
    Represents the :func:`.on_command` event.
    """
    command_completion = "command_completion"
    """Called when a command has completed its invocation.
    Represents the :func:`.on_command_completion` event.
    """
    command_error = "command_error"
    """Called when an error is raised inside a command either through user input error, check failure, or an error in your own code.
    Represents the :func:`.on_command_error` event.
    """
    slash_command = "slash_command"
    """Called when a slash command is found and is about to be invoked.
    Represents the :func:`.on_slash_command` event.
    """
    slash_command_completion = "slash_command_completion"
    """Called when a slash command has completed its invocation.
    Represents the :func:`.on_slash_command_completion` event.
    """
    slash_command_error = "slash_command_error"
    """Called when an error is raised inside a slash command either through user input error, check failure, or an error in your own code.
    Represents the :func:`.on_slash_command_error` event.
    """
    user_command = "user_command"
    """Called when a user command is found and is about to be invoked.
    Represents the :func:`.on_user_command` event.
    """
    user_command_completion = "user_command_completion"
    """Called when a user command has completed its invocation.
    Represents the :func:`.on_user_command_completion` event.
    """
    user_command_error = "user_command_error"
    """Called when an error is raised inside a user command either through check failure, or an error in your own code.
    Represents the :func:`.on_user_command_error` event.
    """
    message_command = "message_command"
    """Called when a message command is found and is about to be invoked.
    Represents the :func:`.on_message_command` event.
    """
    message_command_completion = "message_command_completion"
    """Called when a message command has completed its invocation.
    Represents the :func:`.on_message_command_completion` event.
    """
    message_command_error = "message_command_error"
    """Called when an error is raised inside a message command either through check failure, or an error in your own code.
    Represents the :func:`.on_message_command_error` event.
    """


class ApplicationRoleConnectionMetadataType(Enum):
    """Represents the type of a role connection metadata value.

    These offer comparison operations, which allow guilds to configure role requirements
    based on the metadata value for each user and a guild-specified configured value.

    .. versionadded:: 2.8
    """

    integer_less_than_or_equal = 1
    """The metadata value (``integer``) is less than or equal to the guild's configured value."""
    integer_greater_than_or_equal = 2
    """The metadata value (``integer``) is greater than or equal to the guild's configured value."""
    integer_equal = 3
    """The metadata value (``integer``) is equal to the guild's configured value."""
    integer_not_equal = 4
    """The metadata value (``integer``) is not equal to the guild's configured value."""
    datetime_less_than_or_equal = 5
    """The metadata value (``ISO8601 string``) is less than or equal to the guild's configured value (``integer``; days before current date)."""
    datetime_greater_than_or_equal = 6
    """The metadata value (``ISO8601 string``) is greater than or equal to the guild's configured value (``integer``; days before current date)."""
    boolean_equal = 7
    """The metadata value (``integer``) is equal to the guild's configured value."""
    boolean_not_equal = 8
    """The metadata value (``integer``) is not equal to the guild's configured value."""


class OnboardingPromptType(Enum):
    """Represents the type of onboarding prompt.

    .. versionadded:: 2.9
    """

    multiple_choice = 0
    """The prompt is a multiple choice prompt."""
    dropdown = 1
    """The prompt is a dropdown prompt."""


class SKUType(Enum):
    """Represents the type of an SKU.

    .. versionadded:: 2.10
    """

    durable = 2
    """Represents a durable one-time purchase."""
    consumable = 3
    """Represents a consumable one-time purchase."""
    subscription = 5
    """Represents a recurring subscription."""
    subscription_group = 6
    """Represents a system-generated group for each :attr:`subscription` SKU."""


class EntitlementType(Enum):
    """Represents the type of an entitlement.

    .. versionadded:: 2.10
    """

    purchase = 1
    """Represents an entitlement purchased by a user."""
    premium_subscription = 2
    """Represents an entitlement for a Discord Nitro subscription."""
    developer_gift = 3
    """Represents an entitlement gifted by the application developer."""
    test_mode_purchase = 4
    """Represents an entitlement purchased by a developer in application test mode."""
    free_purchase = 5
    """Represents an entitlement granted when the SKU was free."""
    user_gift = 6
    """Represents an entitlement gifted by another user."""
    premium_purchase = 7
    """Represents an entitlement claimed by a user for free as a Discord Nitro subscriber."""
    application_subscription = 8
    """Represents an entitlement for an application subscription."""


class SubscriptionStatus(Enum):
    """Represents the status of a subscription.

    .. versionadded:: 2.10
    """

    active = 0
    """Represents an active Subscription which is scheduled to renew."""
    ending = 1
    """Represents an active Subscription which will not renew."""
    inactive = 2
    """Represents an inactive Subscription which is not being charged."""


class PollLayoutType(Enum):
    """Specifies the layout of a :class:`Poll`.

    .. versionadded:: 2.10
    """

    default = 1
    """The default poll layout type."""


class VoiceChannelEffectAnimationType(Enum):
    """The type of an emoji reaction effect animation in a voice channel.

    .. versionadded:: 2.10
    """

    premium = 0
    """A fun animation, sent by a Nitro subscriber."""
    basic = 1
    """A standard animation."""


class MessageReferenceType(Enum):
    """Specifies the type of :class:`MessageReference`. This can be used to determine
    if a message is e.g. a reply or a forwarded message.

    .. versionadded:: 2.10
    """

    default = 0
    """A standard message reference used in message replies."""
    forward = 1
    """Reference used to point to a message at a point in time (forward)."""


T = TypeVar("T")


def create_unknown_value(cls: Type[T], val: Any) -> T:
    value_cls = cls._enum_value_cls_  # type: ignore
    name = f"unknown_{val}"
    return value_cls(name=name, value=val)


def try_enum(cls: Type[T], val: Any) -> T:
    """A function that tries to turn the value into enum ``cls``.

    If it fails it returns a proxy invalid value instead.
    """
    try:
        return cls._enum_value_map_[val]  # type: ignore
    except (KeyError, TypeError, AttributeError):
        return create_unknown_value(cls, val)


def enum_if_int(cls: Type[T], val: Any) -> T:
    """A function that tries to turn the value into enum ``cls``.

    If it fails it returns a proxy invalid value instead.
    """
    if not isinstance(val, int):
        return val
    return try_enum(cls, val)


def try_enum_to_int(val: Any) -> Any:
    if isinstance(val, int):
        return val
    try:
        return val.value
    except Exception:
        return val



# SPDX-License-Identifier: MIT

from typing import List, Optional, TypedDict

from .snowflake import Snowflake
from .user import User


class WidgetChannel(TypedDict):
    id: Snowflake
    name: str
    position: int


class WidgetActivity(TypedDict):
    name: str


class WidgetMember(User, total=False):
    # `activity` is used starting api v8, `game` is used in older versions.
    # Since widgets are sometimes used with the unversioned URL, we support both
    # as long as v6 is still the default.
    activity: WidgetActivity
    game: WidgetActivity
    status: str
    avatar_url: str
    deaf: bool
    self_deaf: bool
    mute: bool
    self_mute: bool
    suppress: bool


class Widget(TypedDict):
    id: Snowflake
    name: str
    instant_invite: Optional[str]
    channels: List[WidgetChannel]
    members: List[WidgetMember]
    presence_count: int


class WidgetSettings(TypedDict):
    enabled: bool
    channel_id: Optional[Snowflake]



# SPDX-License-Identifier: MIT

from typing import Optional, TypedDict

from .snowflake import Snowflake, SnowflakeList
from .user import User


class PartialEmoji(TypedDict):
    id: Optional[Snowflake]
    name: Optional[str]


class Emoji(PartialEmoji, total=False):
    roles: SnowflakeList
    user: User
    require_colons: bool
    managed: bool
    animated: bool
    available: bool


class EditEmoji(TypedDict):
    name: str
    roles: Optional[SnowflakeList]



# SPDX-License-Identifier: MIT

from __future__ import annotations

from typing import List, Optional, TypedDict

from typing_extensions import NotRequired

from .snowflake import Snowflake
from .user import User


class PartialSoundboardSound(TypedDict):
    sound_id: Snowflake
    volume: float


class SoundboardSound(PartialSoundboardSound):
    name: str
    emoji_id: Optional[Snowflake]
    emoji_name: Optional[str]
    available: bool


class GuildSoundboardSound(SoundboardSound):
    guild_id: NotRequired[Snowflake]
    user: NotRequired[User]  # only available via REST, given appropriate permissions


class ListGuildSoundboardSounds(TypedDict):
    items: List[GuildSoundboardSound]



# SPDX-License-Identifier: MIT

from typing import Optional, TypedDict

from typing_extensions import NotRequired

from .snowflake import SnowflakeList
from .user import AvatarDecorationData, User


class BaseMember(TypedDict):
    nick: NotRequired[Optional[str]]
    avatar: NotRequired[Optional[str]]
    roles: SnowflakeList
    joined_at: str
    premium_since: NotRequired[Optional[str]]
    deaf: bool
    mute: bool
    pending: NotRequired[bool]
    permissions: NotRequired[str]
    communication_disabled_until: NotRequired[Optional[str]]
    flags: int
    avatar_decoration_data: NotRequired[Optional[AvatarDecorationData]]


class Member(BaseMember, total=False):
    user: User


class MemberWithUser(BaseMember):
    user: User


class UserWithMember(User, total=False):
    member: BaseMember



# SPDX-License-Identifier: MIT

from __future__ import annotations

import asyncio
import datetime
from typing import (
    TYPE_CHECKING,
    Any,
    AsyncIterator,
    Awaitable,
    Callable,
    Dict,
    List,
    Optional,
    TypeVar,
    Union,
    cast,
)

from .app_commands import application_command_factory
from .audit_logs import AuditLogEntry
from .automod import AutoModRule
from .bans import BanEntry
from .entitlement import Entitlement
from .errors import NoMoreItems
from .guild_scheduled_event import GuildScheduledEvent
from .integrations import PartialIntegration
from .object import Object
from .subscription import Subscription
from .threads import Thread
from .utils import maybe_coroutine, snowflake_time, time_snowflake

__all__ = (
    "ReactionIterator",
    "HistoryIterator",
    "BanIterator",
    "AuditLogIterator",
    "GuildIterator",
    "MemberIterator",
    "GuildScheduledEventUserIterator",
    "EntitlementIterator",
    "SubscriptionIterator",
    "PollAnswerIterator",
)

if TYPE_CHECKING:
    from .abc import Messageable, Snowflake
    from .app_commands import APIApplicationCommand
    from .guild import Guild
    from .member import Member
    from .message import Message
    from .state import ConnectionState
    from .types.audit_log import (
        AuditLog as AuditLogPayload,
        AuditLogEntry as AuditLogEntryPayload,
        AuditLogEvent,
    )
    from .types.entitlement import Entitlement as EntitlementPayload
    from .types.guild import Ban as BanPayload, Guild as GuildPayload
    from .types.guild_scheduled_event import (
        GuildScheduledEventUser as GuildScheduledEventUserPayload,
    )
    from .types.message import Message as MessagePayload
    from .types.subscription import Subscription as SubscriptionPayload
    from .types.threads import Thread as ThreadPayload
    from .types.user import PartialUser as PartialUserPayload
    from .user import User

T = TypeVar("T")
OT = TypeVar("OT")
_Func = Callable[[T], Union[OT, Awaitable[OT]]]

OLDEST_OBJECT = Object(id=0)


class _AsyncIterator(AsyncIterator[T]):
    __slots__ = ()

    async def next(self) -> T:
        raise NotImplementedError

    def get(self, **attrs: Any) -> Awaitable[Optional[T]]:
        def predicate(elem: T) -> bool:
            for attr, val in attrs.items():
                nested = attr.split("__")
                obj = elem
                for attribute in nested:
                    obj = getattr(obj, attribute)

                if obj != val:
                    return False
            return True

        return self.find(predicate)

    async def find(self, predicate: _Func[T, bool]) -> Optional[T]:
        while True:
            try:
                elem = await self.next()
            except NoMoreItems:
                return None

            ret = await maybe_coroutine(predicate, elem)
            if ret:
                return elem

    def chunk(self, max_size: int) -> _ChunkedAsyncIterator[T]:
        if max_size <= 0:
            raise ValueError("async iterator chunk sizes must be greater than 0.")
        return _ChunkedAsyncIterator(self, max_size)

    def map(self, func: _Func[T, OT]) -> _MappedAsyncIterator[OT]:
        return _MappedAsyncIterator(self, func)

    def filter(self, predicate: Optional[_Func[T, bool]]) -> _FilteredAsyncIterator[T]:
        return _FilteredAsyncIterator(self, predicate)

    async def flatten(self) -> List[T]:
        return [element async for element in self]

    async def __anext__(self) -> T:
        try:
            return await self.next()
        except NoMoreItems:
            raise StopAsyncIteration from None


class _ChunkedAsyncIterator(_AsyncIterator[List[T]]):
    def __init__(self, iterator: _AsyncIterator[T], max_size: int) -> None:
        self.iterator = iterator
        self.max_size = max_size

    async def next(self) -> List[T]:
        ret: List[T] = []
        n = 0
        while n < self.max_size:
            try:
                item = await self.iterator.next()
            except NoMoreItems:
                if ret:
                    return ret
                raise
            else:
                ret.append(item)
                n += 1
        return ret


class _MappedAsyncIterator(_AsyncIterator[OT]):
    def __init__(self, iterator: _AsyncIterator[T], func: _Func[T, OT]) -> None:
        self.iterator = iterator
        self.func = func

    async def next(self) -> OT:
        # this raises NoMoreItems and will propagate appropriately
        item = await self.iterator.next()
        return await maybe_coroutine(self.func, item)


class _FilteredAsyncIterator(_AsyncIterator[T]):
    def __init__(self, iterator: _AsyncIterator[T], predicate: Optional[_Func[T, bool]]) -> None:
        self.iterator = iterator

        if predicate is None:
            predicate = bool  # similar to the `filter` builtin, a `None` filter drops falsy items

        self.predicate: _Func[T, bool] = predicate

    async def next(self) -> T:
        getter = self.iterator.next
        pred = self.predicate
        while True:
            # propagate NoMoreItems similar to _MappedAsyncIterator
            item = await getter()
            ret = await maybe_coroutine(pred, item)
            if ret:
                return item


class ReactionIterator(_AsyncIterator[Union["User", "Member"]]):
    def __init__(self, message, emoji, limit: int = 100, after=None) -> None:
        self.message = message
        self.limit = limit
        self.after = after
        state = message._state
        self.getter = state.http.get_reaction_users
        self.state = state
        self.emoji = emoji
        self.guild = message.guild
        self.channel_id = message.channel.id
        self.users = asyncio.Queue()

    async def next(self) -> Union[User, Member]:
        if self.users.empty():
            await self.fill_users()

        try:
            return self.users.get_nowait()
        except asyncio.QueueEmpty:
            raise NoMoreItems from None

    async def fill_users(self) -> None:
        if self.limit > 0:
            retrieve = min(self.limit, 100)

            after = self.after.id if self.after else None
            data: List[PartialUserPayload] = await self.getter(
                self.channel_id, self.message.id, self.emoji, retrieve, after=after
            )

            if data:
                self.limit -= retrieve
                self.after = Object(id=int(data[-1]["id"]))

            for element in reversed(data):
                if self.guild is None or isinstance(self.guild, Object):
                    await self.users.put(self.state.create_user(data=element))
                else:
                    member_id = int(element["id"])
                    member = self.guild.get_member(member_id)
                    if member is not None:
                        await self.users.put(member)
                    else:
                        await self.users.put(self.state.create_user(data=element))


class HistoryIterator(_AsyncIterator["Message"]):
    """Iterator for receiving a channel's message history.

    The messages endpoint has two behaviours we care about here:
    If ``before`` is specified, the messages endpoint returns the `limit`
    newest messages before ``before``, sorted with newest first. For filling over
    100 messages, update the ``before`` parameter to the oldest message received.
    Messages will be returned in order by time.
    If ``after`` is specified, it returns the ``limit`` oldest messages after
    ``after``, sorted with newest first. For filling over 100 messages, update the
    ``after`` parameter to the newest message received. If messages are not
    reversed, they will be out of order (99-0, 199-100, so on)

    A note that if both ``before`` and ``after`` are specified, ``before`` is ignored by the
    messages endpoint.

    Parameters
    ----------
    messageable: :class:`abc.Messageable`
        Messageable class to retrieve message history from.
    limit: :class:`int`
        Maximum number of messages to retrieve
    before: Optional[Union[:class:`abc.Snowflake`, :class:`datetime.datetime`]]
        Message before which all messages must be.
    after: Optional[Union[:class:`abc.Snowflake`, :class:`datetime.datetime`]]
        Message after which all messages must be.
    around: Optional[Union[:class:`abc.Snowflake`, :class:`datetime.datetime`]]
        Message around which all messages must be. Limit max 101. Note that if
        limit is an even number, this will return at most limit+1 messages.
    oldest_first: Optional[:class:`bool`]
        If set to ``True``, return messages in oldest->newest order. Defaults to
        ``True`` if `after` is specified, otherwise ``False``.
    """

    def __init__(
        self,
        messageable: Messageable,
        limit: Optional[int] = 100,
        before: Optional[Union[Snowflake, datetime.datetime]] = None,
        after: Optional[Union[Snowflake, datetime.datetime]] = None,
        around: Optional[Union[Snowflake, datetime.datetime]] = None,
        oldest_first: Optional[bool] = None,
    ) -> None:
        if isinstance(before, datetime.datetime):
            before = Object(id=time_snowflake(before, high=False))
        if isinstance(after, datetime.datetime):
            after = Object(id=time_snowflake(after, high=True))
        if isinstance(around, datetime.datetime):
            around = Object(id=time_snowflake(around))

        if oldest_first is None:
            self.reverse = after is not None
        else:
            self.reverse = oldest_first

        self.messageable = messageable
        self.limit = limit
        self.before = before
        self.after = after or OLDEST_OBJECT
        self.around = around

        self._filter: Optional[Callable[[MessagePayload], bool]] = None

        self.state = self.messageable._state
        self.logs_from = self.state.http.logs_from
        self.messages = asyncio.Queue()

        if self.around:
            if self.limit is None:
                raise ValueError("history does not support around with limit=None")
            if self.limit > 101:
                raise ValueError("history max limit 101 when specifying around parameter")
            elif self.limit == 101:
                self.limit = 100  # Thanks Discord

            self._retrieve_messages = self._retrieve_messages_around_strategy
            if self.before and self.after:
                self._filter = lambda m: self.after.id < int(m["id"]) < self.before.id  # type: ignore
            elif self.before:
                self._filter = lambda m: int(m["id"]) < self.before.id  # type: ignore
            elif self.after:
                self._filter = lambda m: self.after.id < int(m["id"])
        else:
            if self.reverse:
                self._retrieve_messages = self._retrieve_messages_after_strategy
                if self.before:
                    self._filter = lambda m: int(m["id"]) < self.before.id  # type: ignore
            else:
                self._retrieve_messages = self._retrieve_messages_before_strategy
                if self.after and self.after != OLDEST_OBJECT:
                    self._filter = lambda m: int(m["id"]) > self.after.id

    async def next(self) -> Message:
        if self.messages.empty():
            await self.fill_messages()

        try:
            return self.messages.get_nowait()
        except asyncio.QueueEmpty:
            raise NoMoreItems from None

    def _get_retrieve(self):
        limit = self.limit
        if limit is None or limit > 100:
            retrieve = 100
        else:
            retrieve = limit
        self.retrieve = retrieve
        return retrieve > 0

    async def fill_messages(self) -> None:
        if not hasattr(self, "channel"):
            # do the required set up
            channel = await self.messageable._get_channel()
            self.channel = channel

        if self._get_retrieve():
            data = await self._retrieve_messages(self.retrieve)
            if len(data) < 100:
                self.limit = 0  # terminate the infinite loop

            if self.reverse:
                data = reversed(data)
            if self._filter:
                data = filter(self._filter, data)

            channel = self.channel
            for element in data:
                await self.messages.put(self.state.create_message(channel=channel, data=element))

    async def _retrieve_messages(self, retrieve) -> List[MessagePayload]:
        """Retrieve messages and update next parameters."""
        raise NotImplementedError

    async def _retrieve_messages_before_strategy(self, retrieve):
        """Retrieve messages using before parameter."""
        before = self.before.id if self.before else None
        data: List[MessagePayload] = await self.logs_from(self.channel.id, retrieve, before=before)
        if len(data):
            if self.limit is not None:
                self.limit -= retrieve
            self.before = Object(id=int(data[-1]["id"]))
        return data

    async def _retrieve_messages_after_strategy(self, retrieve):
        """Retrieve messages using after parameter."""
        after = self.after.id if self.after else None
        data: List[MessagePayload] = await self.logs_from(self.channel.id, retrieve, after=after)
        if len(data):
            if self.limit is not None:
                self.limit -= retrieve
            self.after = Object(id=int(data[0]["id"]))
        return data

    async def _retrieve_messages_around_strategy(self, retrieve):
        """Retrieve messages using around parameter."""
        if self.around:
            around = self.around.id if self.around else None
            data: List[MessagePayload] = await self.logs_from(
                self.channel.id, retrieve, around=around
            )
            self.around = None
            return data
        return []


class BanIterator(_AsyncIterator["BanEntry"]):
    """Iterator for receiving a guild's bans.

    The bans endpoint has two behaviours we care about here:
    If ``before`` is specified, the bans endpoint returns the ``limit``
    bans with user ids before ``before``, sorted with the oldest first. For filling over
    1000 bans, update the ``before`` parameter to the newest user received.
    If ``after`` is specified, it returns the ``limit`` bans with user ids after
    ``after``, sorted with the oldest first. For filling over 1000 bans, update the
    ``after`` parameter to the oldest user received.

    A note that if both ``before`` and ``after`` are specified, ``after`` is ignored by the
    bans endpoint.

    Parameters
    ----------
    guild: :class:`~disnake.Guild`
        The guild to get bans from.
    limit: Optional[:class:`int`]
        Maximum number of bans to retrieve.
    before: Optional[:class:`abc.Snowflake`]
        Object before which all bans must be.
    after: Optional[:class:`abc.Snowflake`]
        Object after which all bans must be.
    """

    def __init__(
        self,
        guild: Guild,
        limit: Optional[int] = None,
        before: Optional[Snowflake] = None,
        after: Optional[Snowflake] = None,
    ) -> None:
        self.guild = guild
        self.limit = limit
        self.before = before
        self.after = after or OLDEST_OBJECT

        self.state = self.guild._state
        self.get_bans = self.state.http.get_bans
        self.bans = asyncio.Queue()

        self._filter: Optional[Callable[[BanPayload], bool]] = None

        if self.before:
            self._retrieve_bans = self._retrieve_bans_before_strategy
            if self.after != OLDEST_OBJECT:
                self._filter = lambda b: int(b["user"]["id"]) > self.after.id
        else:
            self._retrieve_bans = self._retrieve_bans_after_strategy

    async def next(self) -> BanEntry:
        if self.bans.empty():
            await self.fill_bans()

        try:
            return self.bans.get_nowait()
        except asyncio.QueueEmpty:
            raise NoMoreItems from None

    def _get_retrieve(self) -> bool:
        self.retrieve = min(self.limit, 1000) if self.limit is not None else 1000
        return self.retrieve > 0

    async def fill_bans(self) -> None:
        if self._get_retrieve():
            data = await self._retrieve_bans(self.retrieve)
            if len(data) < 1000:
                self.limit = 0  # terminate the infinite loop

            if self._filter:
                data = filter(self._filter, data)

            for element in data:
                await self.bans.put(
                    BanEntry(
                        user=self.state.create_user(data=element["user"]),
                        reason=element["reason"],
                    )
                )

    async def _retrieve_bans_before_strategy(self, retrieve: int) -> List[BanPayload]:
        """Retrieve bans using before parameter."""
        before = self.before.id if self.before else None
        data: List[BanPayload] = await self.get_bans(self.guild.id, retrieve, before=before)
        if len(data):
            if self.limit is not None:
                self.limit -= len(data)
            self.before = Object(id=int(data[0]["user"]["id"]))
        return data

    async def _retrieve_bans_after_strategy(self, retrieve: int) -> List[BanPayload]:
        """Retrieve bans using after parameter."""
        after = self.after.id if self.after else None
        data: List[BanPayload] = await self.get_bans(self.guild.id, retrieve, after=after)
        if len(data):
            if self.limit is not None:
                self.limit -= len(data)
            self.after = Object(id=int(data[-1]["user"]["id"]))
        return data


class AuditLogIterator(_AsyncIterator["AuditLogEntry"]):
    def __init__(
        self,
        guild: Guild,
        limit: Optional[int] = None,
        before: Optional[Union[Snowflake, datetime.datetime]] = None,
        after: Optional[Union[Snowflake, datetime.datetime]] = None,
        user_id: Optional[int] = None,
        action_type: Optional[AuditLogEvent] = None,
        oldest_first: bool = False,
    ) -> None:
        if isinstance(before, datetime.datetime):
            before = Object(id=time_snowflake(before, high=False))
        if isinstance(after, datetime.datetime):
            after = Object(id=time_snowflake(after, high=True))

        self.limit: Optional[int] = limit
        self.before: Optional[Snowflake] = before
        self.after: Snowflake = after or OLDEST_OBJECT
        self.user_id: Optional[int] = user_id
        self.action_type: Optional[AuditLogEvent] = action_type

        self.guild = guild
        self._state = guild._state
        self.request = guild._state.http.get_audit_logs

        self.entries: asyncio.Queue[AuditLogEntry] = asyncio.Queue()

        self._filter: Optional[Callable[[AuditLogEntryPayload], bool]] = None
        if oldest_first:
            self._strategy = self._after_strategy
            if self.before:
                self._filter = lambda m: int(m["id"]) < self.before.id  # type: ignore
        else:
            self._strategy = self._before_strategy
            if self.after and self.after != OLDEST_OBJECT:
                self._filter = lambda m: int(m["id"]) > self.after.id

    async def _before_strategy(self, retrieve: int) -> AuditLogPayload:
        before = self.before.id if self.before else None
        data: AuditLogPayload = await self.request(
            self.guild.id,
            limit=retrieve,
            user_id=self.user_id,
            action_type=self.action_type,
            before=before,
        )

        entries = data.get("audit_log_entries", [])
        if entries:
            if self.limit is not None:
                self.limit -= retrieve
            self.before = Object(id=int(entries[-1]["id"]))
        return data

    async def _after_strategy(self, retrieve: int) -> AuditLogPayload:
        after = self.after.id
        data: AuditLogPayload = await self.request(
            self.guild.id,
            limit=retrieve,
            user_id=self.user_id,
            action_type=self.action_type,
            after=after,
        )

        entries = data.get("audit_log_entries", [])
        if entries:
            if self.limit is not None:
                self.limit -= retrieve
            # note: unlike other paginated endpoints, this one returns entries in ascending
            # order when using `after`; hence `-1` (instead of `0`) is correct here.
            self.after = Object(id=int(entries[-1]["id"]))
        return data

    async def next(self) -> AuditLogEntry:
        if self.entries.empty():
            await self._fill()

        try:
            return self.entries.get_nowait()
        except asyncio.QueueEmpty:
            raise NoMoreItems from None

    def _get_retrieve(self):
        limit = self.limit
        if limit is None or limit > 100:
            retrieve = 100
        else:
            retrieve = limit
        self.retrieve = retrieve
        return retrieve > 0

    async def _fill(self) -> None:
        if self._get_retrieve():
            log_data = await self._strategy(self.retrieve)
            entries = log_data.get("audit_log_entries")
            if len(entries) < 100:
                self.limit = 0  # terminate the infinite loop

            if self._filter:
                entries = filter(self._filter, entries)

            state = self._state

            appcmds: Dict[int, APIApplicationCommand] = {}
            for data in log_data.get("application_commands", []):
                try:
                    cmd = application_command_factory(data)
                except TypeError:
                    pass
                else:
                    appcmds[int(data["id"])] = cmd

            automod_rules = {
                int(data["id"]): AutoModRule(guild=self.guild, data=data)
                for data in log_data.get("auto_moderation_rules", [])
            }

            events = {
                int(data["id"]): GuildScheduledEvent(state=state, data=data)
                for data in log_data.get("guild_scheduled_events", [])
            }

            integrations = {
                int(data["id"]): PartialIntegration(guild=self.guild, data=data)
                for data in log_data.get("integrations", [])
            }

            threads = {
                int(data["id"]): Thread(guild=self.guild, state=state, data=data)
                for data in log_data.get("threads", [])
            }

            users = {int(data["id"]): state.create_user(data) for data in log_data.get("users", [])}

            webhooks = {
                int(data["id"]): state.create_webhook(data) for data in log_data.get("webhooks", [])
            }

            for element in entries:
                # https://github.com/discord/discord-api-docs/issues/5055#issuecomment-1266363766
                if element["action_type"] is None:  # pyright: ignore[reportUnnecessaryComparison]
                    continue

                await self.entries.put(
                    AuditLogEntry(
                        data=element,
                        guild=self.guild,
                        application_commands=appcmds,
                        automod_rules=automod_rules,
                        guild_scheduled_events=events,
                        integrations=integrations,
                        threads=threads,
                        users=users,
                        webhooks=webhooks,
                    )
                )


class GuildIterator(_AsyncIterator["Guild"]):
    """Iterator for receiving the client's guilds.

    The guilds endpoint has the same two behaviours as described
    in :class:`HistoryIterator`:
    If ``before`` is specified, the guilds endpoint returns the ``limit``
    newest guilds before ``before``, sorted with newest first. For filling over
    100 guilds, update the ``before`` parameter to the oldest guild received.
    Guilds will be returned in order by time.
    If `after` is specified, it returns the ``limit`` oldest guilds after ``after``,
    sorted with newest first. For filling over 100 guilds, update the ``after``
    parameter to the newest guild received, If guilds are not reversed, they
    will be out of order (99-0, 199-100, so on)

    Not that if both ``before`` and ``after`` are specified, ``before`` is ignored by the
    guilds endpoint.

    Parameters
    ----------
    bot: :class:`disnake.Client`
        The client to retrieve the guilds from.
    limit: :class:`int`
        Maximum number of guilds to retrieve.
    before: Optional[Union[:class:`abc.Snowflake`, :class:`datetime.datetime`]]
        Object before which all guilds must be.
    after: Optional[Union[:class:`abc.Snowflake`, :class:`datetime.datetime`]]
        Object after which all guilds must be.
    """

    def __init__(
        self, bot, limit: Optional[int], before=None, after=None, with_counts: bool = True
    ) -> None:
        if isinstance(before, datetime.datetime):
            before = Object(id=time_snowflake(before, high=False))
        if isinstance(after, datetime.datetime):
            after = Object(id=time_snowflake(after, high=True))

        self.bot = bot
        self.limit = limit
        self.before = before
        self.after = after
        self.with_counts = with_counts

        self._filter = None

        self.state = self.bot._connection
        self.get_guilds = self.bot.http.get_guilds
        self.guilds = asyncio.Queue()

        if self.before:
            self.reverse = True
            self._retrieve_guilds = self._retrieve_guilds_before_strategy  # type: ignore
            if self.after:
                self._filter = lambda m: int(m["id"]) > self.after.id  # type: ignore
        else:
            self.reverse = False
            self._retrieve_guilds = self._retrieve_guilds_after_strategy  # type: ignore

    async def next(self) -> Guild:
        if self.guilds.empty():
            await self.fill_guilds()

        try:
            return self.guilds.get_nowait()
        except asyncio.QueueEmpty:
            raise NoMoreItems from None

    def _get_retrieve(self):
        limit = self.limit
        if limit is None or limit > 200:
            retrieve = 200
        else:
            retrieve = limit
        self.retrieve = retrieve
        return retrieve > 0

    def create_guild(self, data):
        from .guild import Guild

        return Guild(state=self.state, data=data)

    async def fill_guilds(self) -> None:
        if self._get_retrieve():
            data = await self._retrieve_guilds(self.retrieve)
            if len(data) < 200:
                self.limit = 0

            if self.reverse:
                data = reversed(data)
            if self._filter:
                data = filter(self._filter, data)

            for element in data:
                await self.guilds.put(self.create_guild(element))

    async def _retrieve_guilds(self, retrieve) -> List[Guild]:
        """Retrieve guilds and update next parameters."""
        raise NotImplementedError

    async def _retrieve_guilds_before_strategy(self, retrieve):
        """Retrieve guilds using before parameter."""
        before = self.before.id if self.before else None
        data: List[GuildPayload] = await self.get_guilds(
            retrieve, before=before, with_counts=self.with_counts
        )
        if len(data):
            if self.limit is not None:
                self.limit -= retrieve
            self.before = Object(id=int(data[0]["id"]))
        return data

    async def _retrieve_guilds_after_strategy(self, retrieve):
        """Retrieve guilds using after parameter."""
        after = self.after.id if self.after else None
        data: List[GuildPayload] = await self.get_guilds(
            retrieve, after=after, with_counts=self.with_counts
        )
        if len(data):
            if self.limit is not None:
                self.limit -= retrieve
            self.after = Object(id=int(data[-1]["id"]))
        return data


class MemberIterator(_AsyncIterator["Member"]):
    def __init__(self, guild, limit: Optional[int] = 1000, after=None) -> None:
        if isinstance(after, datetime.datetime):
            after = Object(id=time_snowflake(after, high=True))

        self.guild = guild
        self.limit = limit
        self.after = after or OLDEST_OBJECT

        self.state = self.guild._state
        self.get_members = self.state.http.get_members
        self.members = asyncio.Queue()

    async def next(self) -> Member:
        if self.members.empty():
            await self.fill_members()

        try:
            return self.members.get_nowait()
        except asyncio.QueueEmpty:
            raise NoMoreItems from None

    def _get_retrieve(self):
        limit = self.limit
        if limit is None or limit > 1000:
            retrieve = 1000
        else:
            retrieve = limit
        self.retrieve = retrieve
        return retrieve > 0

    async def fill_members(self) -> None:
        if self._get_retrieve():
            after = self.after.id if self.after else None
            data = await self.get_members(self.guild.id, self.retrieve, after)
            if not data:
                # no data, terminate
                return

            if len(data) < 1000:
                self.limit = 0  # terminate loop

            self.after = Object(id=int(data[-1]["user"]["id"]))

            for element in reversed(data):
                await self.members.put(self.create_member(element))

    def create_member(self, data):
        from .member import Member

        return Member(data=data, guild=self.guild, state=self.state)


class ArchivedThreadIterator(_AsyncIterator["Thread"]):
    def __init__(
        self,
        channel_id: int,
        guild: Guild,
        limit: Optional[int],
        joined: bool,
        private: bool,
        before: Optional[Union[Snowflake, datetime.datetime]] = None,
    ) -> None:
        self.channel_id = channel_id
        self.guild = guild
        self.limit = limit
        self.joined = joined
        self.private = private
        self.http = guild._state.http

        if joined and not private:
            raise ValueError("Cannot iterate over joined public archived threads")

        self.before: Optional[str]
        if before is None:
            self.before = None
        elif isinstance(before, datetime.datetime):
            if joined:
                self.before = str(time_snowflake(before, high=False))
            else:
                self.before = before.isoformat()
        else:
            if joined:
                self.before = str(before.id)
            else:
                self.before = snowflake_time(before.id).isoformat()

        self.update_before: Callable[[ThreadPayload], str] = self.get_archive_timestamp

        if joined:
            self.endpoint = self.http.get_joined_private_archived_threads
            self.update_before = self.get_thread_id
        elif private:
            self.endpoint = self.http.get_private_archived_threads
        else:
            self.endpoint = self.http.get_public_archived_threads

        self.queue: asyncio.Queue[Thread] = asyncio.Queue()
        self.has_more: bool = True

    async def next(self) -> Thread:
        if self.queue.empty():
            await self.fill_queue()

        try:
            return self.queue.get_nowait()
        except asyncio.QueueEmpty:
            raise NoMoreItems from None

    @staticmethod
    def get_archive_timestamp(data: ThreadPayload) -> str:
        return data["thread_metadata"]["archive_timestamp"]

    @staticmethod
    def get_thread_id(data: ThreadPayload) -> str:
        return data["id"]  # type: ignore

    async def fill_queue(self) -> None:
        if not self.has_more:
            raise NoMoreItems

        limit = 50 if self.limit is None else max(self.limit, 50)
        data = await self.endpoint(self.channel_id, before=self.before, limit=limit)

        # This stuff is obviously WIP because 'members' is always empty
        threads: List[ThreadPayload] = data.get("threads", [])
        for d in reversed(threads):
            self.queue.put_nowait(self.create_thread(d))

        self.has_more = data.get("has_more", False)
        if self.limit is not None:
            self.limit -= len(threads)
            if self.limit <= 0:
                self.has_more = False

        if self.has_more:
            self.before = self.update_before(threads[-1])

    def create_thread(self, data: ThreadPayload) -> Thread:
        from .threads import Thread

        return Thread(guild=self.guild, state=self.guild._state, data=data)


class GuildScheduledEventUserIterator(_AsyncIterator[Union["User", "Member"]]):
    def __init__(
        self,
        event: GuildScheduledEvent,
        limit: Optional[int],
        with_members: bool,
        before: Optional[Snowflake],
        after: Optional[Snowflake],
    ) -> None:
        self.event: GuildScheduledEvent = event
        self.limit: Optional[int] = limit
        self.with_members: bool = with_members
        self.before: Optional[Snowflake] = before
        self.after: Optional[Snowflake] = after

        self.state: ConnectionState = event._state
        self.get_event_users = self.state.http.get_guild_scheduled_event_users
        self.users = asyncio.Queue()

        self._filter: Optional[Callable[[GuildScheduledEventUserPayload], bool]] = None
        if self.before is not None:
            self._strategy = self._before_strategy
            if self.after is not None:
                self._filter = lambda u: int(u["user"]["id"]) > cast("Snowflake", self.after).id
            # reverse if using `before` strategy, since chunks are always received in
            # ascending order (200-299, 100-199, 0-99) regardless of before/after
            self.reverse = True
        else:
            self._strategy = self._after_strategy
            self.reverse = False

    async def next(self) -> Union[User, Member]:
        if self.users.empty():
            await self.fill_users()

        try:
            return self.users.get_nowait()
        except asyncio.QueueEmpty:
            raise NoMoreItems from None

    def _get_retrieve(self) -> bool:
        limit = self.limit
        if limit is None or limit > 100:
            retrieve = 100
        else:
            retrieve = limit
        self.retrieve: int = retrieve
        return retrieve > 0

    def create_user(self, data: GuildScheduledEventUserPayload) -> Union[User, Member]:
        from .member import Member

        user_data = data["user"]
        member_data = data.get("member")
        if member_data is not None and (guild := self.event.guild) is not None:
            return guild.get_member(int(user_data["id"])) or Member(
                data=member_data, user_data=user_data, guild=guild, state=self.state
            )
        else:
            return self.state.store_user(data["user"])

    async def fill_users(self) -> None:
        if not self._get_retrieve():
            return

        data = await self._strategy(self.retrieve)
        if len(data) < 100:
            self.limit = 0  # terminate loop

        if self.reverse:
            data = reversed(data)
        if self._filter:
            data = filter(self._filter, data)

        for user in data:
            await self.users.put(self.create_user(user))

    async def _before_strategy(self, retrieve: int) -> List[GuildScheduledEventUserPayload]:
        before = self.before.id if self.before else None
        data = await self.get_event_users(
            self.event.guild_id,
            self.event.id,
            limit=retrieve,
            with_member=self.with_members,
            before=before,
        )

        if len(data):
            if self.limit is not None:
                self.limit -= retrieve
            # users are always returned in ascending order
            self.before = Object(id=int(data[0]["user"]["id"]))
        return data

    async def _after_strategy(self, retrieve: int) -> List[GuildScheduledEventUserPayload]:
        after = self.after.id if self.after else None
        data = await self.get_event_users(
            self.event.guild_id,
            self.event.id,
            limit=retrieve,
            with_member=self.with_members,
            after=after,
        )

        if len(data):
            if self.limit is not None:
                self.limit -= retrieve
            self.after = Object(id=int(data[-1]["user"]["id"]))
        return data


# The endpoint for this paginates like audit logs,
# i.e. descending when no parameter or `before` is given,
# and ascending when `after` is given.
class EntitlementIterator(_AsyncIterator["Entitlement"]):
    def __init__(
        self,
        application_id: int,
        *,
        state: ConnectionState,
        limit: Optional[int],
        user_id: Optional[int] = None,
        guild_id: Optional[int] = None,
        sku_ids: Optional[List[int]] = None,
        before: Optional[Union[Snowflake, datetime.datetime]] = None,
        after: Optional[Union[Snowflake, datetime.datetime]] = None,
        exclude_ended: bool = False,
        exclude_deleted: bool = True,
        oldest_first: bool = False,
    ) -> None:
        if isinstance(before, datetime.datetime):
            before = Object(id=time_snowflake(before, high=False))
        if isinstance(after, datetime.datetime):
            after = Object(id=time_snowflake(after, high=True))

        self.application_id: int = application_id
        self.limit: Optional[int] = limit
        self.before: Optional[Snowflake] = before
        self.after: Snowflake = after or OLDEST_OBJECT
        self.user_id: Optional[int] = user_id
        self.guild_id: Optional[int] = guild_id
        self.sku_ids: Optional[List[int]] = sku_ids
        self.exclude_ended: bool = exclude_ended
        self.exclude_deleted: bool = exclude_deleted

        self.state: ConnectionState = state
        self.request = state.http.get_entitlements

        self.entitlements: asyncio.Queue[Entitlement] = asyncio.Queue()

        self._filter: Optional[Callable[[EntitlementPayload], bool]] = None
        if oldest_first:
            self._strategy = self._after_strategy
            if self.before:
                self._filter = lambda m: int(m["id"]) < self.before.id  # type: ignore
        else:
            self._strategy = self._before_strategy
            if self.after and self.after != OLDEST_OBJECT:
                self._filter = lambda m: int(m["id"]) > self.after.id

    async def next(self) -> Entitlement:
        if self.entitlements.empty():
            await self._fill()

        try:
            return self.entitlements.get_nowait()
        except asyncio.QueueEmpty:
            raise NoMoreItems from None

    def _get_retrieve(self) -> bool:
        limit = self.limit
        if limit is None or limit > 100:
            retrieve = 100
        else:
            retrieve = limit
        self.retrieve: int = retrieve
        return retrieve > 0

    async def _fill(self) -> None:
        if not self._get_retrieve():
            return

        data = await self._strategy(self.retrieve)
        if len(data) < 100:
            self.limit = 0  # terminate loop

        if self._filter:
            data = filter(self._filter, data)

        for entitlement in data:
            await self.entitlements.put(Entitlement(data=entitlement, state=self.state))

    async def _before_strategy(self, retrieve: int) -> List[EntitlementPayload]:
        before = self.before.id if self.before else None
        data = await self.request(
            self.application_id,
            before=before,
            limit=retrieve,
            user_id=self.user_id,
            guild_id=self.guild_id,
            exclude_ended=self.exclude_ended,
            exclude_deleted=self.exclude_deleted,
        )

        if len(data):
            if self.limit is not None:
                self.limit -= retrieve
            self.before = Object(id=int(data[-1]["id"]))
        return data

    async def _after_strategy(self, retrieve: int) -> List[EntitlementPayload]:
        after = self.after.id
        data = await self.request(
            self.application_id,
            after=after,
            limit=retrieve,
            user_id=self.user_id,
            guild_id=self.guild_id,
            exclude_ended=self.exclude_ended,
            exclude_deleted=self.exclude_deleted,
        )

        if len(data):
            if self.limit is not None:
                self.limit -= retrieve
            # endpoint returns items in ascending order when `after` is used
            self.after = Object(id=int(data[-1]["id"]))
        return data


class SubscriptionIterator(_AsyncIterator["Subscription"]):
    def __init__(
        self,
        sku_id: int,
        *,
        state: ConnectionState,
        user_id: Optional[int] = None,  # required, except for oauth queries
        limit: Optional[int] = None,
        before: Optional[Union[Snowflake, datetime.datetime]] = None,
        after: Optional[Union[Snowflake, datetime.datetime]] = None,
    ) -> None:
        if isinstance(before, datetime.datetime):
            before = Object(id=time_snowflake(before, high=False))
        if isinstance(after, datetime.datetime):
            after = Object(id=time_snowflake(after, high=True))

        self.sku_id: int = sku_id
        self.user_id: Optional[int] = user_id
        self.limit: Optional[int] = limit
        self.before: Optional[Snowflake] = before
        self.after: Snowflake = after or OLDEST_OBJECT

        self._state: ConnectionState = state
        self.request = self._state.http.get_subscriptions
        self.subscriptions: asyncio.Queue[Subscription] = asyncio.Queue()

        self._filter: Optional[Callable[[SubscriptionPayload], bool]] = None
        if self.before:
            self._strategy = self._before_strategy
            if self.after != OLDEST_OBJECT:
                self._filter = lambda s: int(s["id"]) > self.after.id
        else:
            self._strategy = self._after_strategy

    async def next(self) -> Subscription:
        if self.subscriptions.empty():
            await self._fill()

        try:
            return self.subscriptions.get_nowait()
        except asyncio.QueueEmpty:
            raise NoMoreItems from None

    def _get_retrieve(self) -> bool:
        limit = self.limit
        if limit is None or limit > 100:
            retrieve = 100
        else:
            retrieve = limit
        self.retrieve: int = retrieve
        return retrieve > 0

    async def _fill(self) -> None:
        if not self._get_retrieve():
            return

        data = await self._strategy(self.retrieve)
        if len(data) < 100:
            self.limit = 0  # terminate loop

        if self._filter:
            data = filter(self._filter, data)

        for subscription in data:
            await self.subscriptions.put(Subscription(data=subscription, state=self._state))

    async def _before_strategy(self, retrieve: int) -> List[SubscriptionPayload]:
        before = self.before.id if self.before else None
        data = await self.request(
            self.sku_id,
            before=before,
            limit=retrieve,
            user_id=self.user_id,
        )

        if len(data):
            if self.limit is not None:
                self.limit -= retrieve
            # since pagination order isn't documented, don't rely on results being sorted one way or the other
            self.before = Object(id=min(int(data[0]["id"]), int(data[-1]["id"])))
        return data

    async def _after_strategy(self, retrieve: int) -> List[SubscriptionPayload]:
        after = self.after.id
        data = await self.request(
            self.sku_id,
            after=after,
            limit=retrieve,
            user_id=self.user_id,
        )

        if len(data):
            if self.limit is not None:
                self.limit -= retrieve
            self.after = Object(id=max(int(data[0]["id"]), int(data[-1]["id"])))
        return data


class PollAnswerIterator(_AsyncIterator[Union["User", "Member"]]):
    def __init__(
        self,
        message: Message,
        answer_id: int,
        *,
        limit: Optional[int],
        after: Optional[Snowflake] = None,
    ) -> None:
        self.channel_id: int = message.channel.id
        self.message_id: int = message.id
        self.answer_id: int = answer_id
        self.guild: Optional[Guild] = message.guild
        self.state: ConnectionState = message._state

        self.limit: Optional[int] = limit
        self.after: Optional[Snowflake] = after

        self.getter = message._state.http.get_poll_answer_voters
        self.users = asyncio.Queue()

    async def next(self) -> Union[User, Member]:
        if self.users.empty():
            await self.fill_users()

        try:
            return self.users.get_nowait()
        except asyncio.QueueEmpty:
            raise NoMoreItems from None

    def _get_retrieve(self) -> bool:
        self.retrieve = min(self.limit, 100) if self.limit is not None else 100
        return self.retrieve > 0

    async def fill_users(self) -> None:
        if self._get_retrieve():
            after = self.after.id if self.after else None
            data = (
                await self.getter(
                    channel_id=self.channel_id,
                    message_id=self.message_id,
                    answer_id=self.answer_id,
                    after=after,
                    limit=self.retrieve,
                )
            )["users"]

            if len(data):
                if self.limit is not None:
                    self.limit -= self.retrieve
                self.after = Object(id=int(data[-1]["id"]))

            if len(data) < 100:
                self.limit = 0  # terminate loop

            for element in data:
                member = None
                if not (self.guild is None or isinstance(self.guild, Object)):
                    member = self.guild.get_member(int(element["id"]))
                await self.users.put(member or self.state.create_user(data=element))
